from catapult.deploy.constants import HOST_WEB, HOST_BOT, HOST_MAINTAIN, HOST_ALL


def phase(paths, storage, _, __, ssh, logger):
    # check date for hosts
    ssh.execute_on_hosts('date')

    # move check file on hosts
    local_path = '{}/check.sh'.format(paths.get('local.release'))
    remote_path = '/tmp/check_{}.sh'.format(storage.get('release.name'))

    ssh.move_on_hosts(
        local_path,
        remote_path
    )

    ssh.execute_on_hosts('chmod +x {}'.format(remote_path))

    is_cluster_mode = storage.is_cluster_mode()
    if is_cluster_mode:
        logger.info('check cluster servers')

        hosts = storage.get_hosts(HOST_ALL)
        for host in hosts:
            logger.info('start checking cluster host "{}"'.format(host))
            ssh.execute_on_host(host, '{} all'.format(remote_path))
            logger.success('checking cluster host "{}" completed'.format(host))

        return

    logger.info('check cloud servers')

    for host_type in (HOST_WEB, HOST_BOT, HOST_MAINTAIN):
        hosts = storage.get_hosts(host_type)

        for host in hosts:
            logger.info('start checking "{}" host "{}"'.format(host_type, host))
            ssh.execute_on_host(
                host,
                '{} {}'.format(
                    remote_path,
                    host_type
                )
            )
            logger.success('checking "{}" host "{}" completed'.format(host_type, host))



import re
from catapult.library.helpers import parse_nested_parameters
from catapult.library.exceptions import InvalidRequestException, ConfigParserException
from catapult.deploy.constants import SERVER_WEB, SERVER_BOT, SERVER_MAINTAIN
from catapult.deploy.constants import TRANSPORT_TELEGRAM, TRANSPORT_SLACK, TRANSPORT_PUSHOVER
from catapult.deploy.constants import CLOUD, CLUSTER


def validate_request(request, services):
    if not request['service']:
        raise InvalidRequestException('Service name must be specified')

    if not request['branch']:
        raise InvalidRequestException('Branch name must be specified')

    if not request['place']:
        raise InvalidRequestException('Place name must be specified')

    for command in request['commands']:
        if not isinstance(command, str):
            raise InvalidRequestException('All commands must be string')

    if request['service'] not in services:
        raise InvalidRequestException('Service with name "{}" does not found'.format(request['service']))

    if request['place'] not in services[request['service']]['places']:
        raise InvalidRequestException(
            'Place "{}" for service "{}" does not found'.format(
                request['place'],
                request['service']
            )
        )


def validate_config(config):
    if 'deploy' not in config or not config['deploy']:
        raise ConfigParserException('You must provide section "deploy" in config.yml file')

    if 'user' not in config['deploy']:
        raise ConfigParserException('You must provide "user" in section "deploy" of config.yml file')

    if 'deploy_path' not in config['deploy']:
        raise ConfigParserException('You must provide "deploy_path" in section "deploy" of config.yml file')

    if 'keep_releases' not in config['deploy']:
        raise ConfigParserException('You must provide "keep_releases" in section "deploy" of config.yml file')

    if 'ssh' not in config['deploy'] or not config['deploy']['ssh']:
        raise ConfigParserException('You must provide "ssh" section in section "deploy" of config.yml file')

    if 'user' not in config['deploy']['ssh']:
        raise ConfigParserException('You must provide "user" in section "ssh" in section "deploy" of config.yml file')

    if 'key_file' not in config['deploy']['ssh']:
        raise ConfigParserException(
            'You must provide "key_file" in section "ssh" in section "deploy" of config.yml file'
        )

    if 'git' not in config['deploy'] or not config['deploy']['git']:
        raise ConfigParserException('You must provide "git" section in section "deploy" of config.yml file')

    if 'key_file' not in config['deploy']['git']:
        raise ConfigParserException(
            'You must provide "key_file" in section "git" in section "deploy" of config.yml file'
        )

    if 'server' not in config or not config['server']:
        raise ConfigParserException('You must provide section "server" in config.yml file')

    if 'token' not in config['server']:
        raise ConfigParserException('You must provide "token" in section "server" of config.yml file')

    if 'workers' not in config['server']:
        raise ConfigParserException('You must provide "workers" in section "server" of config.yml file')

    if 'host' not in config['server']:
        raise ConfigParserException('You must provide "host" in section "server" of config.yml file')

    if 'port' not in config['server']:
        raise ConfigParserException('You must provide "port" in section "server" of config.yml file')

    if 'registry' in config:
        if 'server' not in config['registry']:
            raise ConfigParserException('You must provide "server" in section "registry" of config.yml file')

        if 'username' not in config['registry']:
            raise ConfigParserException('You must provide "username" in section "registry" of config.yml file')

        if 'password' not in config['registry']:
            raise ConfigParserException('You must provide "password" in section "registry" of config.yml file')

    if 'notifications' in config:
        for transport in config['notifications']:
            if 'token' not in config['notifications'][transport]:
                raise ConfigParserException(
                    'You must specify "token" field for notification transport "{}"'.format(transport)
                )

            if 'messages' in config['notifications'][transport]:
                if not isinstance(config['notifications'][transport]['messages'], dict):
                    raise ConfigParserException(
                        'Field "messages" for transport "{}" must be a dict in config.yml file'.format(transport)
                    )


def validate_service(service_name, config):
    pattern_validate = '^[a-z0-9-]+$'
    regular_validate = re.compile(pattern_validate)

    if not regular_validate.match(service_name):
        raise ConfigParserException(
            'Name of service "{}" must match to pattern "{}"'.format(
                service_name,
                pattern_validate
            )
        )

    if 'repo' not in config or not config['repo'] or not isinstance(config['repo'], str):
        raise ConfigParserException('You must specify section "repo" in config.yml of service "{}"'.format(service_name))

    if 'places' not in config or not config['places']:
        raise ConfigParserException('You must specify section "places" in config.yml of service "{}"'.format(service_name))

    if 'linked' in config and not isinstance(config['linked'], list):
        raise ConfigParserException(
            'Section "linked" for service "{}" must be a list'.format(service_name)
        )

    if 'parameters' in config and not isinstance(config['parameters'], dict):
        raise ConfigParserException(
            'Section "parameters" for service "{}" must be a dict'.format(service_name)
        )

    for place_name in config['places']:
        place = config['places'][place_name]

        if 'linked' in place:
            raise ConfigParserException(
                'You must remove section "linked" for place "{}" of service "{}"'.format(place_name, service_name)
            )

        if CLUSTER not in place and CLOUD not in place:
            raise ConfigParserException(
                'You must specify "{}" or "{}" section for place "{}" of service "{}"'.format(
                    CLUSTER,
                    CLOUD,
                    place_name,
                    service_name
                )
            )

        if CLUSTER in place and CLOUD in place:
            raise ConfigParserException(
                'You must specify "{}" or "{}" section for place "{}" of service "{}"'.format(
                    CLUSTER,
                    CLOUD,
                    place_name,
                    service_name
                )
            )

        if CLOUD in place:
            if not isinstance(place[CLOUD], dict):
                raise ConfigParserException(
                    'Section "cloud" for place "{}" of service "{}" must be a dict of'
                    ' "{}", "{}" and "{}" servers'.format(
                        SERVER_WEB,
                        SERVER_BOT,
                        SERVER_MAINTAIN,
                        place_name,
                        service_name
                    )
                )

            if SERVER_WEB not in place[CLOUD] and SERVER_BOT not in place[CLOUD]:
                raise ConfigParserException(
                    'You must specify "{}" or "{}" section for place "{}" of service "{}"'.format(
                        SERVER_WEB,
                        SERVER_BOT,
                        place_name,
                        service_name
                    )
                )

            if SERVER_WEB in place[CLOUD] and not isinstance(place[CLOUD][SERVER_WEB], list):
                raise ConfigParserException(
                    'Section "{}" for place "{}" of service "{}" must be list of servers'.format(
                        SERVER_WEB,
                        place_name,
                        service_name
                    )
                )

            if SERVER_BOT in place[CLOUD] and not isinstance(place[CLOUD][SERVER_BOT], list):
                raise ConfigParserException(
                    'Section "{}" for place "{}" of service "{}" must be list of servers'.format(
                        SERVER_BOT,
                        place_name,
                        service_name
                    )
                )

            if SERVER_MAINTAIN in place[CLOUD] and not isinstance(place[CLOUD][SERVER_MAINTAIN], list):
                raise ConfigParserException(
                    'Section "{}" for place "{}" of service "{}" must be a list of servers'.format(
                        SERVER_MAINTAIN,
                        place_name,
                        service_name,
                    )
                )

            # check unique servers
            servers = []
            for host_type in (SERVER_WEB, SERVER_BOT, SERVER_MAINTAIN):
                if host_type in place[CLOUD]:
                    for item in place[CLOUD][host_type]:
                        if 'host' not in item:
                            raise ConfigParserException(
                                'You must specify "{}" host for place "{}" of service "{}"'.format(
                                    host_type,
                                    place_name,
                                    service_name,
                                )
                            )

                        servers.append(item['host'])
            if sorted(list(set(servers))) != sorted(servers):
                raise ConfigParserException(
                    'All hosts for for place "{}" of service "{}" must be unique'.format(
                        place_name,
                        service_name,
                    )
                )

            # check weight for servers
            if SERVER_BOT in place[CLOUD]:
                is_weight_set = False
                for server in place[CLOUD][SERVER_BOT]:
                    if 'weight' in server:
                        if not isinstance(server['weight'], int):
                            raise ConfigParserException(
                                '"weight" for place "{}" of service "{}" must be integer'.format(
                                    place_name,
                                    service_name,
                                )
                            )

                        is_weight_set = True

                    if 'weight' not in server and is_weight_set is True:
                        raise ConfigParserException(
                            'You must specify "weight" for host "{}" in place "{}"'.format(
                                server['host'],
                                place_name,
                            )
                        )

                    if is_weight_set is True and (server['weight'] > 3 or server['weight'] <= 0):
                        raise ConfigParserException(
                            '"weight" for host "{}" in place "{}" must be between 1 and 3'.format(
                                server['host'],
                                place_name,
                            )
                        )

            # check parameters
            is_set_checked = False
            checked_parameters = []
            checked_host = ''
            for host_type in (SERVER_WEB, SERVER_BOT, SERVER_MAINTAIN):
                if host_type in place[CLOUD]:
                    for server in place[CLOUD][host_type]:
                        if 'parameters' in server and not isinstance(server['parameters'], dict):
                            raise ConfigParserException(
                                'Section "parameters" for service "{}" and place "{}" '
                                'and servers type "{}" and host "{}" must be a dict'.format(
                                    service_name,
                                    place_name,
                                    host_type,
                                    server['host']
                                )
                            )

                        parameters = {}
                        if 'parameters' in server:
                            parameters = server['parameters']

                        parameters = parse_nested_parameters(parameters)

                        if not is_set_checked:
                            is_set_checked = True
                            checked_parameters = parameters
                            checked_host = server['host']

                            continue

                        more_diff = list(set(checked_parameters) - set(parameters))
                        less_diff = list(set(parameters) - set(checked_parameters))

                        if more_diff:
                            raise ConfigParserException(
                                'You must specify parameters "{}" for service "{}" and place "{}" '
                                'and servers type "{}" and host "{}"'.format(
                                    ', '.join(more_diff),
                                    service_name,
                                    place_name,
                                    host_type,
                                    server['host']
                                )
                            )

                        if less_diff:
                            raise ConfigParserException(
                                'You must specify parameters "{}" for service "{}" and place "{}" '
                                'and servers type "{}" and host "{}"'.format(
                                    ', '.join(less_diff),
                                    service_name,
                                    place_name,
                                    host_type,
                                    checked_host
                                )
                            )

        if CLUSTER in place:
            if not isinstance(place[CLUSTER], list):
                raise ConfigParserException(
                    'Section "cluster" for place "{}" of service "{}" must be a list of servers'.format(
                        place_name,
                        service_name
                    )
                )

            # check unique servers
            servers = []
            for server in place[CLUSTER]:
                if 'host' not in server:
                    raise ConfigParserException(
                        'You must specify host for place "{}" of service "{}"'.format(
                            place_name,
                            service_name,
                        )
                    )

                servers.append(server['host'])
            if sorted(list(set(servers))) != sorted(servers):
                raise ConfigParserException(
                    'All hosts for for place "{}" of service "{}" must be unique'.format(
                        place_name,
                        service_name,
                    )
                )

            # check weight for servers
            is_weight_set = False
            for server in place[CLUSTER]:
                if 'weight' in server:
                    if not isinstance(server['weight'], int):
                        raise ConfigParserException(
                            '"weight" for place "{}" of service "{}" must be integer'.format(
                                place_name,
                                service_name,
                            )
                        )

                    is_weight_set = True

                if 'weight' not in server and is_weight_set is True:
                    raise ConfigParserException(
                        'You must specify "weight" for host "{}" in place "{}"'.format(
                            server['host'],
                            place_name,
                        )
                    )

                if is_weight_set is True and (server['weight'] > 3 or server['weight'] <= 0):
                    raise ConfigParserException(
                        '"weight" for host "{}" in place "{}" must be between 1 and 3'.format(
                            server['host'],
                            place_name,
                        )
                    )

            # check parameters
            is_set_checked = False
            checked_parameters = []
            checked_host = ''
            for server in place[CLUSTER]:
                if 'parameters' in server and not isinstance(server['parameters'], dict):
                    raise ConfigParserException(
                        'Section "parameters" for service "{}" and place "{}" and host "{}" must be a dict'.format(
                            service_name,
                            place_name,
                            server['host']
                        )
                    )

                parameters = {}
                if 'parameters' in server:
                    parameters = server['parameters']

                parameters = parse_nested_parameters(parameters)

                if not is_set_checked:
                    is_set_checked = True
                    checked_parameters = parameters
                    checked_host = server['host']

                    continue

                more_diff = list(set(checked_parameters) - set(parameters))
                less_diff = list(set(parameters) - set(checked_parameters))

                if more_diff:
                    raise ConfigParserException(
                        'You must specify parameters "{}" for service "{}" and place "{}" '
                        'and host "{}"'.format(
                            ', '.join(more_diff),
                            service_name,
                            place_name,
                            server['host']
                        )
                    )

                if less_diff:
                    raise ConfigParserException(
                        'You must specify parameters "{}" for service "{}" and place "{}" '
                        'and host "{}"'.format(
                            ', '.join(less_diff),
                            service_name,
                            place_name,
                            checked_host
                        )
                    )

        if 'recipes' in place and not isinstance(place['recipes'], list):
            raise ConfigParserException(
                'Section "recipes" for place "{}" of service "{}" must be a list'.format(place_name, service_name)
            )

        if 'parameters' in place and not isinstance(place['parameters'], dict):
            raise ConfigParserException(
                'Section "parameters" for place "{}" of service "{}" must be a dict'.format(place_name, service_name)
            )

        if 'notifications' in place:
            if not isinstance(place['notifications'], dict):
                raise ConfigParserException(
                    'Section "notifications" for place "{}" of service "{}" must be a dict'.format(place_name, service_name)
                )

            for transport in place['notifications']:
                if not isinstance(place['notifications'][transport], dict):
                    raise ConfigParserException(
                        'Section "{}" for place "{}" of service "{}" must be a dict'.format(
                            transport,
                            place_name,
                            service_name
                        )
                    )

                if transport == TRANSPORT_TELEGRAM:
                    if 'chat_id' not in place['notifications'][transport]:
                        raise ConfigParserException(
                            'You must specify "chat_id" field for transport "{}" for place "{}" of service "{}"'.format(
                                transport,
                                place_name,
                                service_name,
                            )
                        )

                    if not isinstance(place['notifications'][transport]['chat_id'], str):
                        raise ConfigParserException(
                            'Field "chat_id" for transport "{}" for place "{}" of service "{}" must be a string'.format(
                                transport,
                                place_name,
                                service_name
                            )
                        )

                if transport == TRANSPORT_SLACK:
                    if 'channel' not in place['notifications'][transport]:
                        raise ConfigParserException(
                            'You must specify "channel" field for transport "{}" for place "{}" of service "{}"'.format(
                                transport,
                                place_name,
                                service_name,
                            )
                        )

                    if not isinstance(place['notifications'][transport]['channel'], str):
                        raise ConfigParserException(
                            'Field "channel" for transport "{}" for place "{}" of service "{}" must be a string'.format(
                                transport,
                                place_name,
                                service_name
                            )
                        )

                if transport == TRANSPORT_PUSHOVER:
                    if 'clients' not in place['notifications'][transport]:
                        raise ConfigParserException(
                            'You must specify "clients" field for transport "{}" for place "{}" of service "{}"'.format(
                                transport,
                                place_name,
                                service_name,
                            )
                        )

                    if not isinstance(place['notifications'][transport]['clients'], list):
                        raise ConfigParserException(
                            'Field "clients" for transport "{}" must be a list for place "{}" of service "{}"'.format(
                                TRANSPORT_PUSHOVER,
                                place_name,
                                service_name,
                            )
                        )



class InvalidRequestException(Exception):
    pass


class ConfigParserException(Exception):
    pass


class BlockerException(Exception):
    pass


class SshException(Exception):
    pass


class InvalidParameterException(Exception):
    pass


class LockException(Exception):
    pass


class ModuleException(Exception):
    pass


class RecipeException(Exception):
    pass


class DeployException(Exception):
    pass



import os
from catapult.library.helpers import parse_yaml
from catapult.library.exceptions import ConfigParserException
from catapult.library.validators import validate_service, validate_config
from catapult.library.transformers import transform_service, transform_config


def parse_services(project_path, main_config):
    services_path = os.path.join(project_path, 'services')

    if not os.path.isdir(services_path):
        raise ConfigParserException('You must provide service folder in path {}'.format(project_path))

    services = {}

    for service in os.listdir(services_path):
        service_path = os.path.join(services_path, service)

        if not os.path.isdir(service_path):
            continue

        config_path = os.path.join(service_path, 'config.yml')

        if not os.path.isfile(config_path):
            raise ConfigParserException('You must specify config.yml file for service {}'.format(service))

        config = parse_yaml(config_path)

        validate_service(service, config)

        services[service] = transform_service(config, main_config)

    return services


def parse_config(project_path):
    config_path = os.path.join(project_path, 'config.yml')

    if not os.path.isfile(config_path):
        raise ConfigParserException('You must provide config.yml file in path {}'.format(project_path))

    config = parse_yaml(config_path)

    validate_config(config)

    transform_config(project_path, config)

    return config




import sqlite3
from json import dumps, loads


class Repository:
    def __init__(self, sqlite_path):
        self.sqlite_path = sqlite_path

    def initialize(self):
        db = sqlite3.connect(self.sqlite_path)
        cursor = db.cursor()
        cursor.execute('''
            CREATE TABLE deploys(
                name TEXT PRIMARY KEY,
                request TEXT,
                status TEXT,
                logger TEXT,
                time_end TEXT
            )
        ''')
        db.commit()

    def get_release(self, release_name):
        db = sqlite3.connect(self.sqlite_path)
        cursor = db.cursor()

        cursor.execute(
            '''
              SELECT
                name,
                request,
                status,
                logger,
                time_end
              FROM
                deploys
              WHERE
                name = ?
          ''',
            (release_name,)
        )
        release = cursor.fetchone()
        db.close()

        if not release:
            return None

        return {
            'name': release[0],
            'request': loads(release[1]),
            'status': release[2],
            'logger': loads(release[3]),
            'time_end': release[4]
        }

    def insert_release(self, name, request, status):
        db = sqlite3.connect(self.sqlite_path)
        cursor = db.cursor()

        cursor.execute(
            '''
                INSERT INTO deploys
                (name, request, status, logger)
                VALUES
                (?, ?, ?, ?)
            ''',
            (
                name, dumps(request), status, dumps([]),
            )
        )
        db.commit()

        db.close()

    def update_release(self, name, status, logger, time_end):
        logger = dumps(logger)

        db = sqlite3.connect(self.sqlite_path)

        cursor = db.cursor()
        cursor.execute(
            '''
                UPDATE deploys SET
                  status = ?,
                  logger = ?,
                  time_end = ?
                WHERE
                  name = ?
            ''',
            (
                status,
                logger,
                time_end,
                name,
            )
        )
        db.commit()

        db.close()



from catapult.deploy.constants import HOST_MAINTAIN
from catapult.library.maintain import execute_maintain


def phase(paths, storage, _, __, ssh, logger):
    hosts = storage.get_hosts(HOST_MAINTAIN)
    for host in hosts:
        ssh.execute_on_host(
            host,
            '{}/install.sh maintain'.format(paths.get('remote.release'))
        )

    logger.success('release installed on maintained servers')

    logger.info('start executing maintain commands')

    execute_maintain(
        storage,
        paths,
        ssh,
        storage.get('release.commands'),
        logger
    )

    logger.success('executing maintain commands completed')



import math
from operator import itemgetter
from catapult.modules.library import calculate_total_weight


class ServerNode:
    def __init__(self, host, capacity):
        self.host = host
        self.capacity = capacity
        self.next_node = None
        self.works = []

    def is_allow_weight(self, weight):
        if weight <= self.capacity:
            return True

        return False

    def minus_weight(self, weight):
        self.capacity -= weight

    def set_next(self, next_node):
        self.next_node = next_node

    def get_next(self):
        return self.next_node

    def get_host(self):
        return self.host

    def append_work(self, work):
        self.works.append(work)

    def get_works(self):
        return self.works


def get_circled_linked_list(servers, point):
    last_node = None
    first_node = None

    for server in servers:
        capacity = math.ceil(server["weight"] * point)

        if capacity < 3:
            capacity = 3

        node = ServerNode(server["host"], capacity)

        if not first_node:
            first_node = node

        if last_node:
            last_node.set_next(node)

        last_node = node

    last_node.set_next(first_node)

    return first_node


def fill_balancing_nodes(works, server_node):
    works = sorted(works, key=itemgetter('weight'), reverse=True)

    current_node = server_node

    for work in works:
        while True:
            if current_node.is_allow_weight(work['weight']):
                current_node.append_work(work)
                current_node.minus_weight(work['weight'])
                current_node = current_node.get_next()
                break

            current_node = current_node.get_next()


def get_balancing(server_node):
    balancing = {}

    current_node = server_node
    first_node_host = current_node.get_host()
    is_first_iteration = True
    while True:
        if first_node_host == current_node.get_host() and not is_first_iteration:
            break

        balancing[current_node.get_host()] = current_node.get_works()
        current_node = current_node.get_next()

        is_first_iteration = False

    return balancing


def balanced_by_host(elements, servers):
    balancing_weight = calculate_total_weight(elements)
    servers_weight = calculate_total_weight(servers)

    point = math.ceil(balancing_weight / float(servers_weight))

    server_node = get_circled_linked_list(servers, point)
    fill_balancing_nodes(elements, server_node)

    return get_balancing(server_node)



import sys
import importlib
import subprocess
import inspect
from os.path import exists
from catapult.library.renders import render_string
from catapult.library.maintain import execute_maintain
from catapult.library.exceptions import RecipeException
from catapult.deploy.bullet import Bullet
from catapult.deploy.constants import STATE_BEFORE, STATE_AFTER
from catapult.deploy.constants import BULLET_LOCAL, BULLET_MAINTAIN, BULLET_CONSOLE, BULLET_REMOTE


class Recipes:
    def __init__(self, storage, paths, ssh, logger):
        self.storage = storage
        self.paths = paths
        self.ssh = ssh
        self.logger = logger

        self.recipes = {}

        recipes = self.storage.get('place.recipes')

        if not recipes:
            return

        sys.path.insert(0, paths.get('local.project_path'))

        for recipe in recipes:
            recipe_path = '{}/{}.py'.format(
                paths.get('local.recipes'),
                recipe
            )

            if not exists(recipe_path):
                raise RecipeException(
                    'Recipe "{}" does not found by path "{}"'.format(
                        recipe,
                        recipe_path
                    )
                )

            imported_module = importlib.import_module('recipes.{}'.format(recipe))

            recipe_arguments = inspect.getfullargspec(imported_module.Recipe.__init__)
            if len(recipe_arguments.args) < 2:
                raise RecipeException(
                    'You must specify second argument of __init__ method of recipe "{}" as Storage object'.format(
                        recipe
                    )
                )

            instance = imported_module.Recipe(
                self.storage,
                self.paths
            )

            self.recipes[recipe] = instance

    def execute(self, state, phase):
        if not self.recipes:
            self.logger.info('recipes list is empty, skipped'.format(phase))

            return

        if state not in (STATE_BEFORE, STATE_AFTER):
            raise RecipeException('Invalid state "{}" for execute recipe'.format(state))

        for recipe_name, recipe_instance in self.recipes.items():
            method_name = '{}_{}'.format(state, phase)

            if method_name not in dir(recipe_instance):
                self.logger.info(
                    'action "{}" phase "{}" for recipe "{}" not exists, skipped'.format(
                        state,
                        phase,
                        recipe_name
                    )
                )

                continue

            bullet = Bullet()

            method_link = getattr(recipe_instance, method_name)
            method_link(bullet)

            commands = bullet.get_commands()

            for command in commands:
                if command['type'] == BULLET_LOCAL:
                    execute_local(
                        self.storage,
                        self.paths,
                        self.logger,
                        recipe_name,
                        command['command']
                    )
                    continue

                if command['type'] == BULLET_REMOTE:
                    execute_remote(
                        self.storage,
                        self.paths,
                        self.ssh,
                        self.logger,
                        recipe_name,
                        command['command'],
                        command['types'],
                        command['is_sudo']
                    )

                    continue

                if command['type'] == BULLET_MAINTAIN:
                    execute_maintain(
                        self.storage,
                        self.paths,
                        self.ssh,
                        command['commands'],
                        self.logger,
                        recipe_name,
                    )

                    continue

                if command['type'] == BULLET_CONSOLE:
                    self.logger.info(command['message'], recipe_name)

                    continue


def execute_local(storage, paths, logger, recipe_name, command):
    command = render_string(command, {
        'storage': storage.get,
        'paths': paths.get,
    })

    command = 'cd {} && {}'.format(
        paths.get('local.code'),
        command
    )

    # show log message
    logger.info('execute local command "{}"'.format(command), recipe_name)

    process = subprocess.Popen(
        command,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True
    )

    for line in iter(process.stdout.readline, ''):
        logger.success(line.strip('\n'), recipe_name)

    for line in iter(process.stderr.readline, ''):
        logger.error(line.strip('\n'), recipe_name)

    process.stdout.close()
    process.stderr.close()

    exit_code = process.wait()
    if exit_code != 0:
        raise RecipeException('Execution error with command "{}"'.format(command))


def execute_remote(storage, paths, ssh, logger, recipe_name, command, types, is_sudo):
    if is_sudo:
        command = 'sudo {}'.format(command)

    command = 'cd {} && {}'.format(
        paths.get('remote.code'),
        command
    )

    command = render_string(command, {
        'storage': storage.get,
        'paths': paths.get,
    })

    # show log message
    logger.info('execute remote command "{}"'.format(command), recipe_name)

    for host_type in types:
        hosts = storage.get_hosts(host_type)

        for host in hosts:
            ssh.execute_on_host(host, command)



import os
from jinja2 import Environment, BaseLoader, FileSystemLoader


def render_path(path, context):
    path, filename = os.path.split(path)

    return Environment(
        loader=FileSystemLoader(path)
    ).get_template(filename).render(context)


def render_string(string, context):
    template = Environment(loader=BaseLoader).from_string(string)

    return template.render(context)


def render_shared(template_path, dest, storage, paths, host):
    storage.start_host(host)
    generated = render_path(template_path, {
        'storage': storage.get,
        'paths': paths.get,
    })
    storage.finish_host()

    with open(dest, 'w') as open_file:
        open_file.write(generated)



from string import Template
from catapult.server.constants import WEBSOCKET_PREFIX


def get_ws_url(server_host, server_port):
    return Template('ws://{}:{}/websocket/$release_name{}'.format(
        server_host,
        server_port,
        WEBSOCKET_PREFIX
    ))



from catapult.deploy.constants import HOST_ALL


def phase(paths, storage, _, __, ssh, logger):
    logger.info('start creating folder for release')
    ssh.execute_on_hosts('mkdir -p {}'.format(
        paths.get('remote.release')
    ))
    logger.success('folder "{}" for release created'.format(paths.get('remote.release')))

    hosts = storage.get_hosts(HOST_ALL)

    for host in hosts:
        ssh.move_on_host(
            host,
            '{}/{}.tar.gz'.format(paths.get('local.builds'), host),
            '{}/archive.tar.gz'.format(paths.get('remote.release'))
        )
        logger.success('tar ball catapulted', host)

    logger.info('start extracting tar balls on all hosts')
    ssh.execute_on_hosts('cd {} && tar -xf archive.tar.gz'.format(
        paths.get('remote.release'))
    )
    logger.success('tar balls extracted')

    ssh.execute_on_hosts('rm {}/archive.tar.gz'.format(paths.get('remote.release')))



import json


def success_response(data):
    response = {
        'isSuccess': True,
        'data': data
    }

    return json.dumps(response)


def fail_response(code, message):
    response = {
        'isSuccess': False,
        'error': {
            'code': code,
            'message': message
        }
    }

    return json.dumps(response)


def event_input(event_type, payload={}):
    data = {
        'event': event_type,
        'payload': payload,
    }

    return json.dumps(data)


def event_error(code, message):
    data = {
        'event': 'error',
        'payload': {
            'code': code,
            'message': message,
        },
    }

    return json.dumps(data)



def phase(_, __, locker, blocker, ___, logger):
    # acquire local lock
    logger.info('try to acquire local lock')
    blocker.acquire_lock()
    logger.success('local lock acquired successfully')

    # remote lock
    logger.info('try to acquire remote locks')
    locker.acquire_lock()
    logger.success('remote locks acquired successfully')



import logging
from tornado.web import RequestHandler
from catapult.server.builders import success_response, fail_response
from catapult.server.constants import INVALID_RELEASE
from catapult.server.routes.check_token import check_token


class GetRelease(RequestHandler):
    def initialize(self, token, repository):
        self.token = token
        self.repository = repository

    @check_token
    def get(self, release_name):
        logging.info('get information about release {}'.format(release_name))

        release = self.repository.get_release(release_name)

        if not release:
            self.write(fail_response(INVALID_RELEASE, 'Release with name {} does not found'.format(release_name)))

            return

        self.write(success_response(release))



import logging
from signal import signal, SIGINT, SIGTERM


class GracefulKiller:
    def __init__(self, is_disabled=False):
        self.is_killed = False
        self.is_worked = False

        handler = self.stub_handler if is_disabled else self.signal_handler

        signal(SIGINT, handler)
        signal(SIGTERM, handler)

    def signal_handler(self, signum, _):
        logging.info('handled signal "{}"'.format(signum))

        self.is_killed = True

        if not self.is_worked:
            raise GracefulKillerException()

    def stub_handler(self, signum, _):
        logging.info('handled stub signal "{}"'.format(signum))

    def work_started(self):
        self.is_worked = True

    def work_finished(self):
        self.is_worked = False


class GracefulKillerException(Exception):
    pass



import logging
from tornado.ioloop import IOLoop
from tornado.httpserver import HTTPServer
from tornado.web import Application
from catapult.server.routes.get_release import GetRelease
from catapult.server.routes.post_release import PostRelease
from catapult.server.routes.websocket import WebSocket
from catapult.server.killer import GracefulKiller, GracefulKillerException


def server(config, services, repository, deploy_queue, write_queue):
    try:
        killer = GracefulKiller()

        server_host = config['server']['host']
        server_port = config['server']['port']
        token = config['server']['token']

        application = Application([
            (r'/releases', PostRelease, dict(
                token=token,
                services=services,
                deploy_queue=deploy_queue,
                write_queue=write_queue,
            )),
            (r'/releases/(.*)', GetRelease, dict(
                token=token,
                repository=repository
            )),
            (r'/websocket/(.*)', WebSocket, dict(
                repository=repository
            )),
        ])

        logging.info('trying start server on {}:{}'.format(server_host, server_port))

        http_server = HTTPServer(application)
        http_server.listen(server_port, server_host)

        IOLoop.current().start()
    except GracefulKillerException:
        logging.info('server process shutting down')



import logging
from catapult.server.constants import DEPLOY_PENDING
from catapult.server.killer import GracefulKiller


def writer(repository, write_queue):
    killer = GracefulKiller()

    while True:
        object = write_queue.get()

        if object['status'] == DEPLOY_PENDING:
            logging.info('write release {}'.format(object['name']))

            repository.insert_release(
                object['name'],
                object['request'],
                object['status'],
            )
        else:
            logging.info('update release {}'.format(object['name']))

            repository.update_release(
                object['name'],
                object['status'],
                object['logger'],
                object['time_end'],
            )




from os import makedirs, remove
from os.path import exists
from catapult.library.exceptions import BlockerException


class Blocker:
    def __init__(self, releases_folder, release_place, release_name):
        self.release_name = release_name

        if not exists(releases_folder):
            makedirs(releases_folder)

        self.lock_file = '{}/.{}.lock'.format(
            releases_folder,
            release_place
        )

    def acquire_lock(self):
        if exists(self.lock_file):
            raise BlockerException(
                'Can not obtain local lock by path "{}"'.format(self.lock_file)
            )

        with open(self.lock_file, 'w') as lock_file:
            lock_file.write(self.release_name)

    def release_lock(self):
        try:
            remove(self.lock_file)
        except:
            pass




DEPLOY_PENDING = 'pending'
DEPLOY_SUCCESS = 'success'
DEPLOY_FAIL = 'fail'

INVALID_JSON = 'INVALID_JSON'
INVALID_REQUEST = 'INVALID_REQUEST'
INVALID_TOKEN = 'INVALID_TOKEN'
INVALID_RELEASE = 'INVALID_RELEASE'
RELEASE_DEPLOYED = 'RELEASE_DEPLOYED'
READ_ONLY = 'READ_ONLY'

INPUT_BEGIN = 'begin'
INPUT_END = 'end'
INPUT_MESSAGE = 'message'

WEBSOCKET_PREFIX = '/input'



import time


def phase(_, storage, locker, blocker, ssh, logger):
    # release remote lock
    logger.info('try to release remote locks')
    locker.release_lock()
    logger.success('remote lock released successfully')

    # local local lock
    logger.info('try to release local lock')
    blocker.release_lock()
    logger.success('local lock released successfully')

    # close ssh connections
    logger.info('try to close all ssh connections')
    ssh.close()
    logger.success('ssh connections closed successfully')

    time_spend = round(time.time() - storage.get('release.time_start'), 2)
    logger.info('deployed in {}s'.format(time_spend))



import logging
from json import loads
from tornado.websocket import WebSocketHandler
from catapult.server.constants import INPUT_BEGIN, INPUT_END, INPUT_MESSAGE, WEBSOCKET_PREFIX
from catapult.server.constants import DEPLOY_PENDING
from catapult.server.constants import INVALID_RELEASE, RELEASE_DEPLOYED, INVALID_JSON, READ_ONLY, INVALID_REQUEST
from catapult.server.builders import event_error

output_connections = {}
logger = {}


class WebSocket(WebSocketHandler):
    def __init__(self, *args, **kwargs):
        super(WebSocket, self).__init__(*args, **kwargs)

        self.is_input = False
        self.release_name = None

    def initialize(self, repository):
        self.repository = repository

    def check_origin(self, origin):
        return True

    def open(self, release_name):
        if release_name.endswith(WEBSOCKET_PREFIX):
            release_name = release_name[0: len(release_name) - len(WEBSOCKET_PREFIX)]
            self.is_input = True

        self.release_name = release_name

        release = self.repository.get_release(release_name)

        if not release:
            self.write_message(event_error(
                INVALID_RELEASE,
                'Release with name "{}" does not found'.format(release_name)
            ))

            self.close()

            return

        if release['status'] != DEPLOY_PENDING:
            self.write_message(event_error(
                RELEASE_DEPLOYED,
                'Release with name "{}" already deployed'.format(release_name)
            ))

            self.close()

            return

        add_connection(self)

        logging.info('open websocket with "{}" and is_input={}'.format(
            self.release_name,
            self.is_input
        ))

    def on_message(self, message):
        if not self.is_input:
            self.write_message(event_error(
                READ_ONLY,
                'This websocket only for read'
            ))

            return

        try:
            message = loads(message)
        except:
            self.write_message(event_error(
                INVALID_JSON,
                'An error occurred with parsing json'
            ))

            return

        if 'event' not in message:
            self.write_message(event_error(
                INVALID_REQUEST,
                'Field event not specified'
            ))

            return

        if message['event'] == INPUT_BEGIN:
            logger[self.release_name] = []

        if message['event'] == INPUT_MESSAGE:
            logger[self.release_name].append(message['payload'])

            send_messages(self.release_name, message['payload'])

        if message['event'] == INPUT_END:
            close_connections(self.release_name)

            del logger[self.release_name]

            self.close()

    def on_close(self):
        if not self.is_input:
            remove_connection(self)

        logging.info('close websocket with "{}" and is_input={}'.format(
            self.release_name,
            self.is_input
        ))


def send_messages(release_name, message):
    if release_name in output_connections:
        for connection in output_connections[release_name]:
            connection.write_message(message)


def close_connections(release_name):
    if release_name in output_connections:
        for connection in output_connections[release_name]:
            connection.close()

        del output_connections[release_name]


def add_connection(self):
    if self.is_input:
        return

    if self.release_name not in output_connections:
        output_connections[self.release_name] = []

    output_connections[self.release_name].append(self)

    if self.release_name in logger:
        for message in logger[self.release_name]:
            self.write_message(message)


def remove_connection(self):
    if self.release_name not in output_connections:
        return

    output_connections[self.release_name].remove(self)



import docker
from catapult.deploy.constants import REGISTER_CHECK, REGISTER_INSTALL, REGISTER_WEB_AND_BOT
from catapult.modules.normalizers import normalize_docker
from catapult.library.exceptions import ModuleException
from catapult.library.renders import render_path


class Module:
    def __init__(self, module_config, storage, paths, register, logger):
        if not storage.get('place.registry'):
            raise ModuleException(
                'You must specify section "registry" in config.yml of service "{}"'.format(
                    storage.get('release.service')
                )
            )

        if 'publish' not in module_config:
            module_config['publish'] = []

        if not isinstance(module_config['publish'], list):
            module_config['publish'] = [module_config['publish']]

        self.module_config = module_config
        self.storage = storage
        self.paths = paths
        self.logger = logger
        self.register = register

        self.tag = "{}:{}".format(storage.get('place.registry'), storage.get('release.name'))

        register(REGISTER_CHECK, REGISTER_WEB_AND_BOT, self.render_check())

    def render_check(self):
        path = '{}/docker/check.j2'.format(
            self.paths.get('catapult.templates'),
        )

        generated = render_path(path, {
            'storage': self.storage.get,
            'paths': self.paths.get,
        })

        return generated

    def configure(self, config):
        config = normalize_docker(config)

        self.register(REGISTER_INSTALL, REGISTER_WEB_AND_BOT, self.render_install(config['workdir']))

        build_image(
            config,
            self.tag,
            self.paths,
            self.storage,
            self.logger
        )

    def render_install(self, workdir):
        path = '{}/docker/install.j2'.format(
            self.paths.get('catapult.templates'),
        )

        volumes = render_volumes(
            workdir,
            self.paths,
            self.storage
        )

        publish = self.render_publish()

        container = 'catapult-{}-{}'.format(
            self.storage.get('release.service'),
            self.storage.get('release.place')
        )

        generated = render_path(path, {
            'container': container,
            'tag': self.tag,
            'server': self.storage.get('config.registry.server'),
            'username': self.storage.get('config.registry.username'),
            'password': self.storage.get('config.registry.password'),
            'volumes': volumes,
            'publish': publish,
        })

        return generated

    def render_publish(self):
        if not self.module_config['publish']:
            self.logger.info('docker mapping ports skipped')

            return ''

        self.logger.info('docker using mapping ports {}'.format(', '.join(self.module_config['publish'])))

        path_publish = '{}/docker/publish.j2'.format(
            self.paths.get('catapult.templates'),
        )

        publications = []

        for publish in self.module_config['publish']:
            generated = render_path(path_publish, {
                'publish': publish,
            })

            publications.append(generated)

        return ' '.join(publications)


def render_volumes(workdir, paths, storage):
    path_volume = '{}/docker/volume.j2'.format(
        paths.get('catapult.templates'),
    )

    shared_folder = "{}/{}/shared".format(
        storage.get('config.deploy.deploy_path'),
        storage.get('release.service')
    )

    volumes = []

    for linked in storage.get('place.linked'):
        from_path = "{}/{}".format(
            shared_folder,
            linked
        )

        to_path = "{}/{}".format(
            workdir.rstrip('/'),
            linked
        )

        generated = render_path(path_volume, {
            'from': from_path,
            'to': to_path,
        })

        volumes.append(generated)

    return '\n'.join(volumes)

def build_image(config, tag, paths, storage, logger):
    client = docker.DockerClient(base_url=storage.get('config.registry.base_url'))

    logger.info('building docker image with tag %s' % tag)

    buildargs = {
        **storage.get('parameters'),
        **config['buildargs']
    }

    try:
        generator = client.images.build(
            path=paths.get('local.code'),
            dockerfile=config['dockerfile'],
            tag=tag,
            nocache=False,
            forcerm=True,
            buildargs=buildargs,
            rm=True
        )
    except docker.errors.BuildError as e:
        for line in e.build_log:
            if 'stream' in line:
                logger.error(line['stream'].strip())

        raise ModuleException('Some error occurred while building docker image')

    logger.info('docker image successfully built, pushing to remote registry')

    auth_config = {
        'username': storage.get('config.registry.username'),
        'password': storage.get('config.registry.password'),
    }

    pushing_stream = client.images.push(
        tag,
        auth_config=auth_config,
        stream=True,
        decode=True
    )

    for line in pushing_stream:
        if 'status' in line:
            logger.info(line['status'].strip())


from json import dumps
from catapult.library.renders import render_path, render_string
from catapult.library.helpers import get_hosts
from catapult.modules.normalizers import normalize_pm2
from catapult.modules.library import get_applications_by_type, get_mirroring_applications, get_balancing_applications
from catapult.modules.balancer import balanced_by_host
from catapult.deploy.constants import SERVER_WEB, SERVER_BOT, HOST_ALL
from catapult.deploy.constants import REGISTER_CHECK, REGISTER_INSTALL, REGISTER_WEB_AND_BOT


class Module:
    def __init__(self, module_config, storage, paths, register, logger):
        self.module_config = module_config
        self.storage = storage
        self.paths = paths
        self.logger = logger

        register(REGISTER_INSTALL, REGISTER_WEB_AND_BOT, self.render_install())
        register(REGISTER_CHECK, REGISTER_WEB_AND_BOT, self.render_check())

    def render_check(self):
        path = '{}/pm2/check.j2'.format(
            self.paths.get('catapult.templates'),
        )

        generated = render_path(path, {
            'storage': self.storage.get,
            'paths': self.paths.get,
        })

        return generated

    def render_install(self):
        path = '{}/pm2/install.j2'.format(
            self.paths.get('catapult.templates'),
        )

        generated = render_path(path, {
            'release_path': self.paths.get('remote.release'),
            'user': self.storage.get('config.deploy.user'),
            'service_name': 'catapult-{}'.format(self.storage.get('release.service')),
            'storage': self.storage.get,
            'paths': self.paths.get,
        })

        return generated

    def configure(self, config):
        config = normalize_pm2(self.storage.get('release.service'), config)

        pm2 = get_pm2(
            config,
            self.storage,
            self.logger
        )

        configs = {}
        for host in pm2:
            self.storage.start_host(host)

            generated = generate_apps(
                pm2[host],
                self.storage,
                self.paths
            )

            configs[host] = get_json(generated)

            self.storage.finish_host()

        return configs


def get_pm2(applications, storage, logger):
    pm2 = {}

    if storage.is_cluster_mode():
        hosts = storage.get_hosts(HOST_ALL)

        for host in hosts:
            pm2[host] = []

            for application in applications:
                pm2[host].append(application)

        return pm2

    web_servers = storage.get_servers(SERVER_WEB)
    bot_servers = storage.get_servers(SERVER_BOT)

    if web_servers:
        logger.info('configuring pm2 process file for web hosts: {}'.format(
            ', '.join(get_hosts(web_servers)))
        )

    if bot_servers:
        logger.info('configuring pm2 process file for bot hosts: {}'.format(
            ', '.join(get_hosts(web_servers)))
        )

    web_applications = get_applications_by_type(applications, SERVER_WEB)
    bot_applications = get_applications_by_type(applications, SERVER_BOT)

    balancing_applications = get_balancing_applications(bot_applications)
    mirroring_applications = get_mirroring_applications(bot_applications)

    balanced = balanced_by_host(balancing_applications, bot_servers)

    # compose applications by hosts
    for server in bot_servers:
        if server['host'] not in pm2:
            pm2[server['host']] = []

        for application in mirroring_applications:
            pm2[server['host']].append(application)

        for application in balanced[server['host']]:
            pm2[server['host']].append(application)

    # compose applications by hosts
    for server in web_servers:
        if server['host'] not in pm2:
            pm2[server['host']] = []

        for application in web_applications:
            pm2[server['host']].append(application)

    return pm2


def generate_apps(applications, storage, paths):
    apps = []

    for application in applications:
        command = render_string(application['command'], {
            'release_path': paths.get('remote.code'),
            'storage': storage.get,
            'paths': paths.get,
        })

        generated = {
            'name': application['name'],
            'script': command,
            'instances': application['instances'],
            'exec_mode': application['mode'],
            'cwd': paths.get('remote.code')
        }

        apps.append(generated)

    return apps


def get_json(applications):
    json = {
        'apps': applications
    }

    return dumps(json)



from os.path import exists
from catapult.deploy.constants import SERVER_WEB
from catapult.deploy.constants import REGISTER_CHECK, REGISTER_INSTALL, REGISTER_WEB
from catapult.modules.normalizers import normalize_nginx
from catapult.library.exceptions import ModuleException
from catapult.library.renders import render_path, render_string
from catapult.library.helpers import get_hosts


class Module:
    def __init__(self, module_config, storage, paths, register, logger):
        if 'template' not in module_config:
            raise ModuleException('You must set "template" parameter for module "nginx"!')

        self.module_config = module_config
        self.storage = storage
        self.paths = paths
        self.logger = logger

        register(REGISTER_CHECK, REGISTER_WEB, self.render_check_web())
        register(REGISTER_INSTALL, REGISTER_WEB, self.render_install_web())

    def render_check_web(self):
        path = '{}/nginx/check.j2'.format(
            self.paths.get('catapult.templates'),
        )

        generated = render_path(path, {
            'storage': self.storage.get,
            'paths': self.paths.get,
        })

        return generated

    def render_install_web(self):
        path = '{}/nginx/install.j2'.format(
            self.paths.get('catapult.templates'),
        )

        generated = render_path(path, {
            'release_path': self.paths.get('remote.release'),
            'file_name': 'catapult-{}'.format(self.storage.get('release.service')),
            'storage': self.storage.get,
            'paths': self.paths.get,
        })

        return generated

    def configure(self, config):
        config = normalize_nginx(config)
        servers = self.storage.get_servers(SERVER_WEB)

        if not servers or not config:
            self.logger.warning('there are not web servers')

            return

        self.logger.info('configuring web hosts: {}'.format(
            ', '.join(get_hosts(servers)))
        )

        template_path = '{}/{}'.format(
            self.paths.get('local.templates'),
            self.module_config['template']
        )

        if not exists(template_path):
            raise ModuleException(
                'Nginx template by path "{}" does not found'.format(template_path)
            )

        configs = {}
        for server in servers:
            host = server['host']

            self.storage.start_host(host)

            configs[host] = render(
                template_path,
                config,
                self.storage,
                self.paths
            )

            self.storage.finish_host()

        return configs


def render(template_path, config, storage, paths):
    path_parameter = '{}/nginx/parameter.j2'.format(
        paths.get('catapult.templates'),
    )

    path_location = '{}/nginx/location.j2'.format(
        paths.get('catapult.templates'),
    )

    locations = []
    for location in config['locations']:
        parameters = []

        for key, value in location['configs'].items():
            if type(value) is list:
                for item in value:
                    generated = render_path(path_parameter, {
                        'key': key,
                        'value': item
                    })

                    parameters.append(generated)
            else:
                generated = render_path(path_parameter, {
                    'key': key,
                    'value': value
                })

                parameters.append(generated)

        generated = render_path(path_location, {
            'modifier': location['modifier'],
            'match': location['match'],
            'parameters': '\n'.join(parameters),
            'storage': storage.get,
            'paths': paths.get,
        })

        locations.append(generated)

    index_files = config['index_files']
    if type(index_files) is list:
        index_files = ' '.join(index_files)

    root = render_string(config['root'], {
        'release_path': paths.get('remote.code'),
        'storage': storage.get,
        'paths': paths.get,
    })

    item = render_path(template_path, {
        'root': root,
        'index_files': index_files,
        'locations': '\n'.join(locations),
        'storage': storage.get,
        'paths': paths.get,
    })

    path_nginx = '{}/nginx/nginx.j2'.format(
        paths.get('catapult.templates'),
    )

    generated = render_path(path_nginx, {
        'item': item,
        'storage': storage.get,
        'paths': paths.get,
    })

    return generated



import requests


def pushover(api_token, client_token, message):
    requests.post(
        'https://api.pushover.net/1/messages.json',
        data={
            'token': api_token,
            'user': client_token,
            'title': 'Catapult',
            'message': message,
            'sound': 'bugle',
        }
    )


def telegram(api_token, chat_id, message):
    payload = dict(
        chat_id=chat_id,
        text=message
    )

    requests.get(
        'https://api.telegram.org/bot{}/sendMessage'.format(api_token),
        params=payload
    )


def slack(api_token, channel, message):
    requests.post(
        'https://slack.com/api/chat.postMessage',
        data={
            'token': api_token,
            'channel': channel,
            'text': message,
        }
    )



import logging
from websocket import create_connection
from catapult.server.constants import INPUT_BEGIN, INPUT_MESSAGE, INPUT_END
from catapult.server.builders import event_input


class Logger:
    def __init__(self, ws_url):
        self.logger = []
        self.ws = create_connection(ws_url)

    def begin(self):
        send_socket(
            self.ws,
            event_input(
                INPUT_BEGIN,
            )
        )

    def event(self, level, message):
        payload = {
            'level': level,
            'message': message,
        }

        self.logger.append(payload)

        send_socket(
            self.ws,
            event_input(
                INPUT_MESSAGE,
                payload
            )
        )

    def get_logger(self):
        return self.logger

    def end(self):
        send_socket(
            self.ws,
            event_input(
                INPUT_END,
            )
        )


def send_socket(connection, data):
    try:
        connection.send(data)
    except Exception as error:
        logging.error('an error "{}" occurred with sending data to socket "{}"'.format(error, data))



from catapult.server.builders import fail_response
from catapult.server.constants import INVALID_TOKEN


def check_token(f):
    def wrapper(self, *args, **kwargs):
        if 'X-Token' not in self.request.headers:
            self.write(fail_response(
                INVALID_TOKEN,
                'You must specify X-Token header'
            ))

            return

        if self.request.headers['X-Token'] != self.token:
            self.write(fail_response(
                INVALID_TOKEN,
                'This token is incorrect'
            ))

            return

        f(self, *args, **kwargs)

    return wrapper



from catapult.deploy.constants import HOST_WEB, HOST_BOT, HOST_ALL


def phase(paths, storage, _, __, ssh, logger):
    is_cluster_mode = storage.is_cluster_mode()

    if is_cluster_mode:
        logger.info('installing on cluster servers')

        hosts = storage.get_hosts(HOST_ALL)
        for host in hosts:
            logger.info('start installing on cluster host "{}"'.format(host))
            ssh.execute_on_host(
                host,
                '{}/install.sh all'.format(paths.get('remote.release'))
            )
            logger.success('installing on cluster host "{}" completed'.format(host))

        return

    logger.info('installing on cloud servers')

    for host_type in (HOST_BOT, HOST_WEB):
        hosts = storage.get_hosts(host_type)

        for host in hosts:
            logger.info('start installing "{}" host "{}"'.format(host_type, host))
            ssh.execute_on_host(
                host,
                '{}/install.sh {}'.format(
                    paths.get('remote.release'),
                    host_type
                )
            )
            logger.success('installing in "{}" host "{}" completed'.format(host_type, host))



from os import system


def phase(paths, storage, _, __, ssh, logger):
    command_template = 'ls -1t {releases_path}' \
                       ' | tail -n +{keep_releases}' \
                       " | sed 's#/##' |" \
                       ' xargs -I ! rm -rf {releases_path}/!'

    keep_releases = int(storage.get('config.deploy.keep_releases')) + 1

    remote_command = command_template.format(
        releases_path=paths.get('remote.releases_folder'),
        keep_releases=keep_releases,
    )

    ssh.execute_on_hosts(remote_command)

    local_command = command_template.format(
        releases_path=paths.get('local.releases_folder'),
        keep_releases=keep_releases,
    )

    system(local_command)

    logger.success('old releases removed')




import time
from catapult.deploy.constants import TRANSPORT_PUSHOVER, TRANSPORT_SLACK, TRANSPORT_TELEGRAM
from catapult.library.renders import render_string
from catapult.library.transports import pushover, telegram, slack


class Notifier:
    def __init__(self, notifications, request, logger):
        self.notifications = notifications
        self.request = request
        self.logger = logger

    def notify(self, is_success=True):
        if not self.notifications:
            return

        for transport, config in self.notifications.items():
            if not config:
                continue

            message = self.get_message(
                is_success,
                config['messages']['success'],
                config['messages']['fail']
            )

            if transport == TRANSPORT_TELEGRAM:
                try:
                    self.logger.info('try to deliver message with transport Telegram')

                    telegram(
                        config['token'],
                        config['chat_id'],
                        message
                    )

                    self.logger.success('message sent')
                except Exception as error:
                    self.logger.error(error)

            if transport == TRANSPORT_PUSHOVER:
                try:
                    self.logger.info('try to deliver message with transport Pushover')

                    for client in config['clients']:
                        pushover(
                            config['token'],
                            client,
                            message
                        )

                    self.logger.success('message sent')
                except Exception as error:
                    self.logger.error(error)

            if transport == TRANSPORT_SLACK:
                try:
                    self.logger.info('try to deliver message with transport Slack')

                    slack(
                        config['token'],
                        config['channel'],
                        message
                    )

                    self.logger.success('message sent')
                except Exception as error:
                    self.logger.error(error)

    def get_message(self, is_success, success, fail):
        self.request['time'] = round(int(time.time()) - self.request['time_start'], 2)

        template = success if is_success else fail

        return render_string(template, self.request)



import logging
import time
from catapult.loggers.server import Logger
from catapult.server.constants import DEPLOY_SUCCESS, DEPLOY_FAIL
from catapult.server.killer import GracefulKiller, GracefulKillerException


def worker(deploy, deploy_queue, write_queue, ws_template):
    try:
        killer = GracefulKiller()

        while True:
            request = deploy_queue.get()

            killer.work_started()

            logging.info('start execute request {}'.format(request['name']))

            ws_url = ws_template.substitute({
                'release_name': request['name']
            })

            logger = Logger(ws_url)

            logger.begin()

            status = DEPLOY_SUCCESS if deploy(request, logger) is True else DEPLOY_FAIL

            logging.info('request {} ended with status {}'.format(request['name'], status))

            logger.end()

            write_queue.put({
                'name': request['name'],
                'status': status,
                'time_end': int(time.time()),
                'logger': logger.get_logger(),
            })

            killer.work_finished()

            if killer.is_killed:
                break
    except GracefulKillerException:
        logging.info('worker process shutting down')



from catapult.deploy.constants import HOST_MAINTAIN
from catapult.library.renders import render_string


def execute_maintain(storage, paths, ssh, commands, logger, recipe_name=None):
    hosts = storage.get_hosts(HOST_MAINTAIN)

    path_commands = wrap_commands_with_path(storage, paths, commands)

    # list of hosts is empty
    if not hosts and path_commands:
        logger.warning('list of maintain servers is empty. Following commands will be skipped:', recipe_name)

        for command in path_commands:
            if recipe_name:
                logger.warning(command, recipe_name)
            else:
                logger.warning(command)

    execute_commands_on_maintain(
        hosts,
        ssh,
        logger,
        path_commands,
        recipe_name
    )


def execute_commands_on_maintain(hosts, ssh, logger, commands, recipe_name=None):
    hosts_iteration = iter(hosts)
    host = next(hosts_iteration)

    for command in commands:
        if recipe_name:
            logger.info('execute recipe command "{}" on host "{}"'.format(command, host), recipe_name)
        else:
            logger.info('execute command "{}"'.format(command), host)

        ssh.execute_on_host(host, command)

        try:
            host = next(hosts_iteration)
        except StopIteration:
            hosts_iteration = iter(hosts)
            host = next(hosts_iteration)


def wrap_commands_with_path(storage, paths, commands):
    path_commands = []

    for command in commands:
        command = 'cd {} && {}'.format(
            paths.get('remote.code'),
            command
        )

        command = render_string(command, {
            'storage': storage.get,
            'paths': paths.get,
        })

        path_commands.append(command)

    return path_commands



from catapult.library.helpers import get_nested
from catapult.library.exceptions import InvalidParameterException
from catapult.library.helpers import get_hosts, get_parameters_by_host
from catapult.deploy.constants import SERVER_BOT, SERVER_MAINTAIN, SERVER_WEB
from catapult.deploy.constants import HOST_ALL, HOST_WEB, HOST_BOT, HOST_MAINTAIN
from catapult.deploy.constants import CLOUD, CLUSTER

PARAMETERS_PREFIX = 'parameters.'


class Storage:
    def __init__(self, config, place, request):
        parameters = {
            **place['parameters'],
            **request['parameters'],
        }

        self.host_parameters = get_parameters_by_host(place)

        storage = {
            'release': request,
            'parameters': parameters,
            'place': place,
            'config': config,
        }

        self.storage = storage
        self.host = None

    def start_host(self, host):
        self.host = host

    def finish_host(self):
        self.host = None

    def get(self, path, default=None):
        # try to get host parameters
        if path.startswith(PARAMETERS_PREFIX) and self.host:
            try:
                return get_nested(
                    path[len(PARAMETERS_PREFIX):],
                    self.host_parameters[self.host]
                )
            except InvalidParameterException:
                pass

        """
            config.*
            place.*
            parameters.*

            release.name
            release.place
            release.branch
            release.service
            release.force
            release.commands
            release.parameters
            release.time_start
        """
        try:
            return get_nested(path, self.storage)
        except InvalidParameterException as error:
            if default is None:
                raise error

            return default

    def get_servers(self, server_type):
        if server_type not in (SERVER_WEB, SERVER_BOT, SERVER_MAINTAIN):
            raise InvalidParameterException(
                'Invalid server type "{}"'.format(server_type)
            )

        # deploy on cluster
        if CLUSTER in self.storage['place']:
            return self.storage['place'][CLUSTER]

        # deploy on cloud
        if server_type not in self.storage['place'][CLOUD]:
            return []

        return self.storage['place'][CLOUD][server_type]

    def get_hosts(self, host_type):
        if host_type not in (HOST_WEB, HOST_BOT, HOST_MAINTAIN, HOST_ALL):
            raise InvalidParameterException(
                'Invalid host type "{}"'.format(host_type)
            )

        # deploy on cluster
        if CLUSTER in self.storage['place']:
            return get_hosts(self.storage['place'][CLUSTER])

        # deploy on cloud
        iteration_types = [host_type]
        if host_type == HOST_ALL:
            iteration_types = (HOST_WEB, HOST_BOT, HOST_MAINTAIN)

        hosts = []
        for iteration_type in iteration_types:
            if iteration_type not in self.storage['place'][CLOUD]:
                continue

            for server in self.storage['place'][CLOUD][iteration_type]:
                hosts.append(server['host'])

        return hosts

    def is_cluster_mode(self):
        if CLUSTER in self.storage['place']:
            return True

        return False



import os
from catapult.library.renders import render_path
from catapult.library.exceptions import LockException


class Locker:
    def __init__(self, paths, storage, ssh, logger):
        self.paths = paths
        self.storage = storage
        self.ssh = ssh
        self.logger = logger

        self.lock_path = '/tmp/locker_{}.sh'.format(storage.get('release.name'))

    def acquire_lock(self):
        template_path = '{}/core/lock.j2'.format(
            self.paths.get('catapult.templates')
        )

        generated = render_path(template_path, {
            'service_folder': self.paths.get('remote.service_folder'),
            'release_name': self.storage.get('release.name')
        })

        file = '{}/locker.sh'.format(self.paths.get('local.release'))
        with open(file, 'w') as open_file:
            open_file.write(generated)

        os.chmod(file, 0o766)

        self.logger.info('move lock file to all hosts')

        self.ssh.move_on_hosts(
            '{}/locker.sh'.format(self.paths.get('local.release')),
            self.lock_path
        )
        self.ssh.execute_on_hosts('chmod +x {}'.format(self.lock_path))

        self.logger.info('moving lock file to all hosts completed')

        self.logger.info('start executing locking on hosts')

        # if is force deploy
        if self.storage.get('release.force'):
            self.logger.info('release old lock because is force deploy')

            self.release_lock()

        locks = self.ssh.execute_on_hosts('{} acquire_lock'.format(self.lock_path))

        self.logger.info('executing locking on hosts completed')

        for host, result in locks.items():
            if not result:
                raise LockException(
                    'Can not get result about acquire lock. Stdout lock from host "{}" is empty'.format(host)
                )

            if result != 'true':
                raise LockException(
                    'Can not obtain lock on host "{}", result is "{}"'.format(host, result)
                )

        # check equivalent releases on all hosts

        # if is force deploy
        if self.storage.get('release.force'):
            return True

        releases = self.ssh.execute_on_hosts('"{}" get_current_release'.format(self.lock_path))

        release = None
        for host, current_release in releases.items():
            if not current_release:
                raise LockException(
                    'Can not get current_release. Stdout lock from host "{}" is empty.'.format(host)
                )

            if release is None:
                release = current_release

            if release != current_release:
                raise LockException(
                    'Release "{}" not compared for host "{}" release "{}"'.format(release, host, current_release)
                )

    def release_lock(self):
        locks = self.ssh.execute_on_hosts('{} release_lock'.format(self.lock_path))

        for host, result in locks.items():
            if not result:
                raise LockException(
                    'Can not get result about release lock. Stdout lock from host "{}" is empty'.format(host)
                )

            if result != 'true':
                raise LockException(
                    'Can not obtain lock on host "{}", result is "{}"'.format(host, result)
                )

            if not self.storage.get('release.force') and result != 'true':
                raise LockException(
                    'Can not release lock on host "{}", result is "{}"'.format(host, result)
                )



import re
from catapult.library.exceptions import ModuleException
from catapult.modules.constants import BOT_BALANCING, BOT_MIRRORING
from catapult.deploy.constants import SERVER_WEB, SERVER_BOT
from catapult.modules.constants import MODE_CLUSTER, MODE_FORK


def normalize_nginx(config):
    if 'root' not in config:
        raise ModuleException('You must specify section "root" for "nginx" section in deploy.yml')

    if 'index_files' not in config:
        raise ModuleException('You must specify section "index_files" for "nginx" section in deploy.yml')

    if not isinstance(config['index_files'], list):
        raise ModuleException('Section "index_files" for "nginx" section in deploy.yml must be a list')

    if 'locations' not in config:
        config['locations'] = []

    locations = []

    for location in config['locations']:
        if 'match' not in location:
            raise ModuleException('You must specify section "match" for locations in deploy.yml')

        if 'modifier' not in location:
            location['modifier'] = ''

        if 'configs' not in location:
            raise ModuleException('You must specify section "configs" for location "{}" in deploy.yml'.format(
                location['match']
            ))

        if not isinstance(location['configs'], dict):
            raise ModuleException('Section "configs" for "nginx" section in deploy.yml must be dictionary')

        locations.append(location)

    config['locations'] = locations

    return config


def normalize_crons(crons):
    is_weight_set = False
    for cron in crons:
        if 'command' not in cron:
            raise ModuleException('You must specify "command" section for all cron commands in deploy.yml')

        if 'planning' not in cron:
            raise ModuleException('You must specify "planning" section for all cron commands in deploy.yml')

        if not isinstance(cron['planning'], dict):
            raise ModuleException('Section "planning" for cron command in deploy.yml must be dictionary')

        # validate minutes, hours, dom, mon, dow
        for item in ('minutes', 'hours', 'dom', 'mon', 'dow'):
            if item not in cron['planning']:
                raise ModuleException(
                    'You must specify "{}" section for cron command "{}" in deploy.yml'.format(
                        item,
                        cron['command'],
                    )
                )

        # validate weight section
        if 'weight' in cron:
            if not isinstance(cron['weight'], int):
                raise ModuleException(
                    'Section "weight" for cron command "{}" in deploy.yml must be integer'.format(
                        cron['command'],
                    )
                )

            is_weight_set = True

        if 'weight' not in cron and is_weight_set is True:
            raise ModuleException(
                'You must specify "weight" for cron command "{}" in deploy.yml'.format(
                    cron['command'],
                )
            )

        if is_weight_set is True and (cron['weight'] > 3 or cron['weight'] <= 0):
            raise ModuleException(
                'Section "weight" for cron command "{}" in deploy.yml must be between 1 and 3'.format(
                    cron['command'],
                )
            )

        # transform weight section
        if 'weight' not in cron:
            cron['weight'] = 1

    return crons


def normalize_supervisor(demons):
    allowed = (BOT_BALANCING, BOT_MIRRORING)
    pattern_validate = '^[a-zA-Z0-9_:\- ]+$'
    regular_validate = re.compile(pattern_validate)

    pattern_clean = '[^a-z0-9]+'
    regular_clean = re.compile(pattern_clean)

    names = []
    is_weight_set = False
    for demon in demons:
        if 'name' not in demon:
            raise ModuleException('You must specify "name" section for all supervisor demons in deploy.yml')

        if not regular_validate.match(demon['name']):
            raise ModuleException(
                'Name "{}" for supervisor demon must match to pattern "{}" in deploy.yml'.format(
                    demon['name'],
                    pattern_validate
                )
            )

        demon['name'] = demon['name'].lower()
        demon['name'] = regular_clean.sub('_', demon['name'])

        if demon['name'] in names:
            raise ModuleException(
                'Name "{}" for supervisor demon already exists in deploy.yml'.format(
                    demon['name']
                )
            )

        names.append(demon['name'])

        if 'command' not in demon:
            raise ModuleException(
                'You must specify "command" section for supervisor demon "{}" in deploy.yml'.format(
                    demon['name']
                )
            )

        if 'instances' not in demon:
            raise ModuleException(
                'You must specify "instances" section for supervisor demon "{}" in deploy.yml'.format(
                    demon['name']
                )
            )

        if not isinstance(demon['instances'], int):
            raise ModuleException(
                'Section "instances" for supervisor demon "{}" in deploy.yml must be integer'.format(
                    demon['name'],
                )
            )

        if demon['instances'] <= 0:
            raise ModuleException(
                'Section "instances" for supervisor demon "{}" in deploy.yml must be more than 0'.format(
                    demon['name'],
                )
            )

        if 'type' not in demon:
            raise ModuleException(
                'You must specify "type" section for supervisor demon "{}" in deploy.yml'.format(
                    demon['name']
                )
            )

        if demon['type'] not in allowed:
            raise ModuleException(
                'Section "type" for supervisor demon "{}" in deploy.yml must by {}'.format(
                    demon['name'],
                    ' or '.join(allowed)
                )
            )

        # validate weight section
        if 'weight' in demon:
            if not isinstance(demon['weight'], int):
                raise ModuleException(
                    'Section "weight" for supervisor demon "{}" in deploy.yml must be integer'.format(
                        demon['name'],
                    )
                )

            is_weight_set = True

        if 'weight' not in demon and is_weight_set is True:
            raise ModuleException(
                'You must specify "weight" for supervisor demon "{}" in deploy.yml'.format(
                    demon['name'],
                )
            )

        if is_weight_set is True and (demon['weight'] > 3 or demon['weight'] <= 0):
            raise ModuleException(
                'Section "weight" for supervisor demon "{}" in deploy.yml must be between 1 and 3'.format(
                    demon['name'],
                )
            )

        # transform weight section
        if 'weight' not in demon:
            demon['weight'] = 1

    return demons


def normalize_systemd(config):
    if 'execute' not in config:
        raise ModuleException('You must specify section "execute" for "systemd" section in deploy.yml')

    if 'environments' not in config:
        config['environments'] = {}

    if 'environments' in config:
        if not isinstance(config['environments'], dict):
            raise ModuleException('Section "environments" for "systemd" section in deploy.yml must be a dict')

    return config


def normalize_pm2(service_name, applications):
    allowed_modes = (MODE_CLUSTER, MODE_FORK)
    allowed_servers = (SERVER_WEB, SERVER_BOT)
    allowed_types = (BOT_BALANCING, BOT_MIRRORING)
    pattern_validate = '^[a-zA-Z0-9_:\- ]+$'
    regular_validate = re.compile(pattern_validate)

    pattern_clean = '[^a-z0-9]+'
    regular_clean = re.compile(pattern_clean)

    names = []
    is_weight_set = False
    for application in applications:
        if 'name' not in application:
            raise ModuleException('You must specify "name" section for all pm2 applications in deploy.yml')

        if not regular_validate.match(application['name']):
            raise ModuleException(
                'Name "{}" for pm2 application must match to pattern "{}" in deploy.yml'.format(
                    application['name'],
                    pattern_validate
                )
            )

        application['name'] = '{}-{}'.format(
            service_name,
            application['name']
        )
        application['name'] = application['name'].lower()
        application['name'] = regular_clean.sub('_', application['name'])

        if application['name'] in names:
            raise ModuleException(
                'Name "{}" for pm2 application already exists in deploy.yml'.format(
                    application['name']
                )
            )

        names.append(application['name'])

        if 'mode' not in application:
            raise ModuleException(
                'You must specify "mode" section for pm2 application "{}" in deploy.yml'.format(
                    application['name']
                )
            )

        if application['mode'] not in allowed_modes:
            raise ModuleException(
                'Section "mode" for pm2 application "{}" in deploy.yml must by {}'.format(
                    application['name'],
                    ' or '.join(allowed_modes)
                )
            )

        if 'command' not in application:
            raise ModuleException(
                'You must specify "command" section for pm2 application "{}" in deploy.yml'.format(
                    application['name']
                )
            )

        if 'instances' not in application:
            raise ModuleException(
                'You must specify "instances" section for pm2 application "{}" in deploy.yml'.format(
                    application['name']
                )
            )

        if not isinstance(application['instances'], int):
            raise ModuleException(
                'Section "instances" for pm2 application "{}" in deploy.yml must be integer'.format(
                    application['name'],
                )
            )

        if application['instances'] <= 0:
            raise ModuleException(
                'Section "instances" for pm2 application "{}" in deploy.yml must be more than 0'.format(
                    application['name'],
                )
            )

        if 'server' not in application:
            raise ModuleException(
                'You must specify "server" section for pm2 application "{}" in deploy.yml'.format(
                    application['name']
                )
            )

        if application['server'] not in allowed_servers:
            raise ModuleException(
                'Section "server" for pm2 application "{}" in deploy.yml must by {}'.format(
                    application['name'],
                    ' or '.join(allowed_servers)
                )
            )

        if application['server'] == SERVER_BOT:
            if 'type' not in application:
                raise ModuleException(
                    'You must specify "type" section for pm2 bot application "{}" in deploy.yml'.format(
                        application['name'],
                    )
                )

            if application['type'] not in allowed_types:
                raise ModuleException(
                    'Section "type" for pm2 bot application "{}" in deploy.yml must by {}'.format(
                        application['name'],
                        ' or '.join(allowed_types)
                    )
                )

        # validate weight section
        if 'weight' in application:
            if not isinstance(application['weight'], int):
                raise ModuleException(
                    'Section "weight" for pm2 application "{}" in deploy.yml must be integer'.format(
                        application['name'],
                    )
                )

            is_weight_set = True

        if 'weight' not in application and is_weight_set is True:
            raise ModuleException(
                'You must specify "weight" for pm2 application "{}" in deploy.yml'.format(
                    application['name'],
                )
            )

        if is_weight_set is True and (application['weight'] > 3 or application['weight'] <= 0):
            raise ModuleException(
                'Section "weight" for pm2 application "{}" in deploy.yml must be between 1 and 3'.format(
                    application['name'],
                )
            )

        # transform weight section
        if 'weight' not in application:
            application['weight'] = 1

    return applications


def normalize_docker(config):
    if 'dockerfile' not in config:
        raise ModuleException('You must specify section "dockerfile" for "docker" section in deploy.yml')

    if 'workdir' not in config:
        raise ModuleException('You must specify section "workdir" for "docker" section in deploy.yml')

    if 'buildargs' not in config:
        config['buildargs'] = {}

    if 'buildargs' in config:
        if not isinstance(config['buildargs'], dict):
            raise ModuleException('Section "buildargs" for "docker" section in deploy.yml must be a dict')

    return config



from catapult.library.exceptions import RecipeException
from catapult.deploy.constants import HOST_WEB, HOST_BOT, HOST_MAINTAIN, HOST_ALL
from catapult.deploy.constants import BULLET_LOCAL, BULLET_MAINTAIN, BULLET_CONSOLE, BULLET_REMOTE


class Bullet:
    def __init__(self):
        self.commands = []

    def local(self, command):
        self.commands.append({
            'type': BULLET_LOCAL,
            'command': command
        })

    def remote(self, types, command, is_sudo=False):
        if type(types) is str:
            types = [types]

        if not isinstance(command, str):
            raise RecipeException('Recipe command "{}" must be a string'.format(command))

        if not isinstance(types, list):
            raise RecipeException('Recipe server types "{}" must be a string or a list'.format(types))

        for host_type in types:
            if host_type not in (HOST_WEB, HOST_BOT, HOST_MAINTAIN, HOST_ALL):
                raise RecipeException(
                    'Recipe server type "{}" is undefined'.format(command, host_type)
                )

        self.commands.append({
            'type': BULLET_REMOTE,
            'is_sudo': True if is_sudo else False,
            'types': types,
            'command': command
        })

    def maintain(self, commands):
        if not isinstance(commands, list):
            raise RecipeException('Commands "{}" for maintain must be a list'.format(commands))

        self.commands.append({
            'type': BULLET_MAINTAIN,
            'commands': commands
        })

    def console(self, message):
        if not isinstance(message, str):
            raise RecipeException('Console message "{}" for recipe must be a string'.format(message))

        self.commands.append({
            'type': BULLET_CONSOLE,
            'message': message
        })

    def get_commands(self):
        return self.commands



import paramiko
from catapult.library.exceptions import SshException


class Ssh:
    def __init__(self, user, key_file, hosts, logger):
        connections = {}
        for host in hosts:
            try:
                connection = paramiko.SSHClient()
                connection.set_missing_host_key_policy(paramiko.AutoAddPolicy())

                connection.connect(host, username=user, key_filename=key_file)

                connections[host] = connection
            except Exception as error:
                raise SshException(
                    'An error "{}" has occurred while establishing a connection to the server "{}"'.format(error, host)
                )

        self.connections = connections
        self.logger = logger

    def close(self):
        for host, connection in self.connections.items():
            try:
                connection.close()
            except Exception as error:
                raise SshException(
                    'An error "{}" has occurred while closing a connection with server "{}"'.format(error, host)
                )

    def execute_on_hosts(self, command):
        result = {}

        for host, connection in self.connections.items():
            result[host] = self.execute_on_host(host, command)

        return result

    def execute_on_host(self, host, command):
        try:
            connection = self.connections[host]

            stdin, stdout, stderr = connection.exec_command(command, get_pty=True)

            stdout_return_line = None
            for line in iter(stdout.readline, ''):
                # take only last line for return
                stdout_return_line = line.strip('\n\r')

                self.logger.success(stdout_return_line, host)

            for line in iter(stderr.readline, ''):
                self.logger.error(line.strip('\n\r'), host)

            if stdout.channel.recv_exit_status() != 0:
                raise SshException(
                    'An error has occurred while execution command "{}" on host "{}"'.format(command, host)
                )

            return stdout_return_line
        except Exception as error:
            raise SshException('Remote execution error "{}"'.format(error))

    def move_on_hosts(self, local_path, remote_path):
        for host, connection in self.connections.items():
            self.move_on_host(host, local_path, remote_path)

    def move_on_host(self, host, local_path, remote_path):
        try:
            connection = self.connections[host]

            sftp = connection.open_sftp()
            result = sftp.put(local_path, remote_path)

            self.logger.success(result, host)
        except Exception as error:
            self.logger.error(error, host)



from blessings import Terminal
from catapult.deploy.constants import LOGGER_ERROR, LOGGER_INFO, LOGGER_PHASE, LOGGER_SUCCESS, LOGGER_WARNING

terminal = Terminal()


class Logger:
    @staticmethod
    def event(level, message):
        terminal_message = ''

        if level == LOGGER_INFO:
            terminal_message = terminal.cyan(message)

        if level == LOGGER_ERROR:
            terminal_message = terminal.red(message)

        if level == LOGGER_SUCCESS:
            terminal_message = terminal.green(message)

        if level == LOGGER_WARNING:
            terminal_message = terminal.yellow(message)

        if level == LOGGER_PHASE:
            terminal_message = terminal.bold_white(message)

        print(terminal_message)



from random import shuffle
from catapult.modules.constants import BOT_MIRRORING, BOT_BALANCING


def calculate_total_weight(items):
    weight = 0
    for item in items:
        weight += item["weight"]

    return weight


def get_balancing_demons(demons):
    balancing = []
    for demon in demons:
        if demon['type'] != BOT_BALANCING:
            continue

        for _id in range(0, demon['instances']):
            demon['name'] = '{}-{}'.format(demon['name'], _id)

            balancing.append(demon)

    shuffle(balancing)

    return balancing


def get_mirroring_demons(demons):
    mirroring = []
    for demon in demons:
        if demon['type'] != BOT_MIRRORING:
            continue

        mirroring.append(demon)

    return mirroring


def get_balancing_applications(applications):
    balancing = []
    for application in applications:
        if application['type'] != BOT_BALANCING:
            continue

        balancing.append(application)

    shuffle(balancing)

    return balancing


def get_mirroring_applications(applications):
    mirroring = []
    for application in applications:
        if application['type'] != BOT_MIRRORING:
            continue

        mirroring.append(application)

    return mirroring


def get_applications_by_type(applications, server_type):
    result = []
    for application in applications:
        if application['server'] != server_type:
            continue

        result.append(application)

    return result



import time
from os.path import normpath, join
from catapult.library.exceptions import ConfigParserException
from catapult.library.helpers import generate_release_name, get_messages
from catapult.deploy.constants import CLOUD, CLUSTER
from catapult.deploy.constants import SERVER_WEB, SERVER_BOT, SERVER_MAINTAIN


def transform_arguments_parameters(arguments):
    parameters = {}

    for pare in arguments:
        stack = pare.split(':')

        if len(stack) != 2:
            raise ConfigParserException('Parameter must be specify as key:value string')

        parameters[stack[0]] = stack[1]

    return parameters


def saturate_request(request):
    release_name = generate_release_name(request['service'], request['branch'])

    wrapper = {
        'name': release_name,
        'time_start': int(time.time()),
    }

    return {
        **request,
        **wrapper,
    }


def from_args_to_request(arguments):
    parameters = transform_arguments_parameters(arguments.parameter)

    return {
        'branch': arguments.branch,
        'commands': arguments.command,
        'force': arguments.force,
        'parameters': parameters,
        'place': arguments.place,
        'service': arguments.service,
    }


def transform_server_request(request):
    commands = []
    if 'commands' in request:
        commands = request['commands']

    parameters = {}
    if 'parameters' in request:
        parameters = request['parameters']

    force = False
    if 'force' in request:
        force = request['force']

    branch = ''
    if 'branch' in request:
        branch = request['branch']

    service = ''
    if 'service' in request:
        service = request['service']

    place = ''
    if 'place' in request:
        place = request['place']

    return {
        'branch': branch.strip(),
        'commands': commands,
        'force': force,
        'parameters': parameters,
        'place': place.strip(),
        'service': service.strip(),
    }


def transform_service(service, config):
    parameters = {}
    if 'parameters' in service:
        parameters = service['parameters']

    recipes = []
    if 'recipes' in service:
        recipes = service['recipes']

    for place in service['places']:
        # set repo
        service['places'][place]['repo'] = service['repo']

        # set registry
        service['places'][place]['registry'] = ''
        if 'registry' in service:
            service['places'][place]['registry'] = service['registry']

        # set linked paths
        service['places'][place]['linked'] = []
        if 'linked' in service:
            service['places'][place]['linked'] = service['linked']

        # transform recipes
        if 'recipes' not in service['places'][place]:
            service['places'][place]['recipes'] = recipes
        else:
            service['places'][place]['recipes'] = list(set(recipes + service['places'][place]['recipes']))

        # transform parameters
        if 'parameters' not in service['places'][place]:
            service['places'][place]['parameters'] = parameters
        else:
            service['places'][place]['parameters'] = {
                **parameters,
                **service['places'][place]['parameters'],
            }

        # transform modules
        if 'modules' not in service['places'][place]:
            service['places'][place]['modules'] = {}

        if service['places'][place]['modules'] is None:
            service['places'][place]['modules'] = {}

        if 'modules' in service['places'][place]:
            for module_name, module_config in service['places'][place]['modules'].items():
                if module_config is None:
                    service['places'][place]['modules'][module_name] = {}

        # transform weight for bot servers of cloud
        if CLOUD in service['places'][place] and SERVER_BOT in service['places'][place][CLOUD]:
            for server in service['places'][place][CLOUD][SERVER_BOT]:
                if 'weight' not in server:
                    server['weight'] = 1

        # transform weight for servers in cluster
        if CLUSTER in service['places'][place]:
            for server in service['places'][place][CLUSTER]:
                if 'weight' not in server:
                    server['weight'] = 1

        # transform parameters for CLOUD
        if CLOUD in service['places'][place]:
            for host_type in (SERVER_WEB, SERVER_BOT, SERVER_MAINTAIN):
                if host_type in service['places'][place][CLOUD]:
                    for server in service['places'][place][CLOUD][host_type]:
                        if 'parameters' not in server:
                            server['parameters'] = {}

        # transform parameters for CLUSTER
        if CLUSTER in service['places'][place]:
            for server in service['places'][place][CLUSTER]:
                if 'parameters' not in server:
                    server['parameters'] = {}

        # transform notifications
        if 'notifications' in config:
            if 'notifications' not in service['places'][place]:
                service['places'][place]['notifications'] = {}

            # set base parameters for transport
            for transport in config['notifications']:
                if transport not in service['places'][place]['notifications']:
                    continue

                # set token for all transports
                service['places'][place]['notifications'][transport]['token'] = \
                    config['notifications'][transport]['token']

                # patch messages templates for transports
                messages = get_messages(config['notifications'][transport])

                if 'messages' not in service['places'][place]['notifications'][transport]:
                    service['places'][place]['notifications'][transport]['messages'] = messages

                if 'success' not in service['places'][place]['notifications'][transport]['messages']:
                    service['places'][place]['notifications'][transport]['messages']['success'] = \
                        messages['success']

                if 'fail' not in service['places'][place]['notifications'][transport]['messages']:
                    service['places'][place]['notifications'][transport]['messages']['fail'] = \
                        messages['fail']

    return service


def transform_config(project_path, config):
    # generate absolute path to ssh.key_file
    ssh_key_file = normpath(config['deploy']['ssh']['key_file'])
    ssh_key_file = join(project_path, ssh_key_file)
    config['deploy']['ssh']['key_file'] = ssh_key_file

    # generate absolute path to git.key_file
    git_key_file = normpath(config['deploy']['git']['key_file'])
    git_key_file = join(project_path, git_key_file)
    config['deploy']['git']['key_file'] = git_key_file

    if 'registry' in config:
        if 'base_url' not in config['registry']:
            config['registry']['base_url'] = 'unix://var/run/docker.sock'

    return config



import logging
from json import loads
from tornado.web import RequestHandler
from catapult.server.builders import success_response, fail_response
from catapult.server.constants import INVALID_JSON, INVALID_REQUEST, DEPLOY_PENDING
from catapult.library.transformers import saturate_request, transform_server_request
from catapult.library.validators import validate_request
from catapult.library.exceptions import InvalidRequestException
from catapult.server.routes.check_token import check_token


class PostRelease(RequestHandler):
    def initialize(self, token, services, deploy_queue, write_queue):
        self.token = token
        self.services = services
        self.deploy_queue = deploy_queue
        self.write_queue = write_queue

    @check_token
    def post(self):
        try:
            request = loads(self.request.body.decode('utf-8'))
        except Exception as error:
            self.write(fail_response(INVALID_JSON, 'An error occurred with parsing json: {}'.format(error)))
            return

        request = transform_server_request(request)
        request = saturate_request(request)

        try:
            validate_request(request, self.services)
        except InvalidRequestException as error:
            self.write(fail_response(INVALID_REQUEST, str(error)))
            return

        logging.info('handle request {}'.format(request))

        self.write_queue.put({
            'name': request['name'],
            'status': DEPLOY_PENDING,
            'request': request,
        })
        self.deploy_queue.put(request)

        data = {
            'release_name': request['name']
        }

        self.write(success_response(data))



import tarfile
from os import makedirs, walk
from multiprocessing import Process
from catapult.deploy.constants import HOST_ALL


def phase(paths, storage, _, __, ___, logger):
    hosts = storage.get_hosts(HOST_ALL)

    makedirs(paths.get('local.builds'))

    processes = {}

    for host in hosts:
        logger.info('start creating build for host "{}"'.format(host))

        process = Process(target=create_build, args=(host, paths))
        process.start()

        processes[host] = process

    for host in hosts:
        processes[host].join()

        logger.success('build for host "{}" created'.format(host))


def create_build(host, paths):
    path_tar = '{}/{}.tar.gz'.format(
        paths.get('local.builds'),
        host
    )

    path_shared = '{}/{}'.format(
        paths.get('local.shared'),
        host
    )

    tar = tarfile.open(path_tar, 'w:gz')

    tar.add(
        '{}/install.sh'.format(paths.get('local.release')),
        arcname='install.sh'
    )

    tar.add(
        path_shared,
        arcname='shared'
    )

    tar.add(
        paths.get('local.code'),
        arcname='code'
    )

    for root, dirs, files in walk('{}/{}'.format(paths.get('local.configs'), host), topdown=True):
        for file in files:
            tar.add(
                '{}/{}'.format(root, file),
                arcname='configs/{}'.format(file)
            )

    tar.close()



import re
import yaml
import shutil
import codecs
from git import Repo
from os import makedirs, walk
from os.path import isfile, expanduser, exists, isdir, join
from catapult.library.exceptions import DeployException
from catapult.library.renders import render_shared, render_path
from catapult.deploy.modules import Modules
from catapult.deploy.constants import HOST_ALL


def phase(paths, storage, _, __, ___, logger):
    # cloning repo
    logger.info('cloning repo "{}" with key_file by path "{}"'.format(
        storage.get('place.repo'),
        storage.get('config.deploy.git.key_file')
    ))

    git_ssh_identity_file = expanduser(storage.get('config.deploy.git.key_file'))
    git_ssh_cmd = 'ssh -i {}'.format(git_ssh_identity_file)

    Repo.clone_from(
        storage.get('place.repo'),
        paths.get('local.code'),
        branch=storage.get('release.branch'),
        env=dict(GIT_SSH_COMMAND=git_ssh_cmd)
    )

    # remove .git folder
    shutil.rmtree('{}/.git'.format(paths.get('local.code')))

    parse_deploy_yaml(storage, paths, logger)
    configure_shared(storage, paths, logger)
    configure_templates(storage, paths, logger)


def parse_deploy_yaml(storage, paths, logger):
    deploy_yaml = {}

    path = '{}/deploy.yml'.format(paths.get('local.code'))

    if not isfile(path):
        logger.warning('deploy.yml file does not exist, skipped')
    else:
        with open(path) as stream:
            try:
                deploy_yaml = yaml.load(stream, Loader=yaml.Loader)
            except Exception as error:
                raise DeployException(
                    'An error occurred "{}" with parsing deploy.yml file by path "{}"'.format(error, path)
                )

    # configure modules
    modules = Modules(storage, paths, logger)
    modules.configure(deploy_yaml)


def configure_shared(storage, paths, logger):
    for host in storage.get_hosts(HOST_ALL):
        logger.info('configuring shared folder for host "{}"'.format(host))

        path_host = '{}/{}'.format(
            paths.get('local.shared'),
            host
        )

        if not exists(path_host):
            makedirs(path_host)

        path_service = '{}/{}/shared'.format(
            paths.get('local.services'),
            storage.get('release.service')
        )

        if not exists(path_service):
            logger.info('shared folder for service does not exist, skipped')

            continue

        pattern = re.compile(r'\.\[([^]]+)\]')

        for root, dirs, files in walk(path_service, topdown=True):
            path = root.replace(path_service, path_host)

            if not isdir(path):
                makedirs(path)

            for name in files:
                modifiers = re.findall(pattern, name)

                if not is_place_fit(modifiers, storage.get('release.place')):
                    logger.info('shared template file "{}" does not fit'.format(name))

                    continue

                cleaned = name
                for modifier in modifiers:
                    cleaned = cleaned.replace('.[{}]'.format(modifier), '')

                file_src = join(root, name)
                file_dest = join(path, cleaned)

                logger.info('copy shared from "{}" -> to -> "{}" to host "{}"'.format(
                    name,
                    cleaned,
                    host
                ))

                render_shared(
                    file_src,
                    file_dest,
                    storage,
                    paths,
                    host
                )


def configure_templates(storage, paths, logger):
    code_path = paths.get('local.code')

    for root, dirs, files in walk(code_path):
        for name in files:
            extension = name[-9:]

            if extension != '.catapult':
                continue

            cleaned = name[:-9]

            file_src = join(root, name)
            file_dest = join(root, cleaned)

            logger.info('configured template file "{}"'.format(file_src))

            generated = render_path(file_src, {
                'get': storage.get,
                'storage': storage.get,
                'paths': paths.get,
            })

            with codecs.open(file_dest, 'w', encoding='utf8') as open_file:
                open_file.write(generated)


def is_place_fit(modifiers, place):
    if len(modifiers) == 0:
        return True

    if place in modifiers:
        return True

    is_consist_not = False
    for modifier in modifiers:
        if modifier[0] == '!':
            is_consist_not = True

            if modifier == '!' + place:
                return False

    return is_consist_not



from catapult.deploy.constants import REGISTER_CHECK, REGISTER_INSTALL, REGISTER_BOT
from catapult.deploy.constants import SERVER_BOT
from catapult.library.renders import render_path, render_string
from catapult.library.helpers import get_hosts
from catapult.modules.library import get_balancing_demons, get_mirroring_demons
from catapult.modules.normalizers import normalize_supervisor
from catapult.modules.constants import BOT_MIRRORING
from catapult.modules.balancer import balanced_by_host


class Module:
    def __init__(self, module_config, storage, paths, register, logger):
        self.module_config = module_config
        self.storage = storage
        self.paths = paths
        self.logger = logger

        register(REGISTER_CHECK, REGISTER_BOT, self.render_check_bot())
        register(REGISTER_INSTALL, REGISTER_BOT, self.render_install_bot())

    def render_check_bot(self):
        path = '{}/supervisor/check.j2'.format(
            self.paths.get('catapult.templates'),
        )

        generated = render_path(path, {
            'storage': self.storage.get,
            'paths': self.paths.get
        })

        return generated

    def render_install_bot(self):
        path = '{}/supervisor/install.j2'.format(
            self.paths.get('catapult.templates'),
        )

        generated = render_path(path, {
            'release_path': self.paths.get('remote.release'),
            'file_name': 'catapult-{}'.format(self.storage.get('release.service')),
            'service_name': self.storage.get('release.service'),
            'storage': self.storage.get,
            'paths': self.paths.get
        })

        return generated

    def configure(self, config):
        demons = normalize_supervisor(config)
        servers = self.storage.get_servers(SERVER_BOT)

        if not servers or not config:
            self.logger.warning('there are not bot servers')

            return

        self.logger.info('configuring bot hosts: {}'.format(
            ', '.join(get_hosts(servers)))
        )

        supervisor = get_supervisor(
            demons,
            servers,
            self.storage.is_cluster_mode()
        )

        configs = {}
        for host in supervisor:
            self.storage.start_host(host)

            configs[host] = render(
                supervisor[host],
                self.storage,
                self.paths
            )

            self.storage.finish_host()

        return configs


def get_supervisor(demons, servers, is_cluster_mode):
    supervisor = {}

    balancing_demons = get_balancing_demons(demons)
    mirroring_demons = get_mirroring_demons(demons)

    # cluster mode
    if is_cluster_mode:
        for server in servers:
            supervisor[server['host']] = []

            for demon in mirroring_demons:
                supervisor[server['host']].append(demon)

            for demon in balancing_demons:
                supervisor[server['host']].append(demon)

        return supervisor

    # cloud mode
    balanced = balanced_by_host(balancing_demons, servers)

    # compose demons by hosts
    for server in servers:
        supervisor[server['host']] = []

        for demon in mirroring_demons:
            supervisor[server['host']].append(demon)

        for demon in balanced[server['host']]:
            supervisor[server['host']].append(demon)

    return supervisor


def render(demons, storage, paths):
    items = []

    path_item = '{}/supervisor/item.j2'.format(
        paths.get('catapult.templates'),
    )

    for demon in demons:
        command = render_string(demon['command'], {
            'release_path': paths.get('remote.code'),
            'storage': storage.get,
            'paths': paths.get
        })

        numprocs = 1
        if demon['type'] == BOT_MIRRORING:
            numprocs = demon['instances']

        item = render_path(path_item, {
            'program': demon['name'],
            'process_name': '%(program_name)s_%(process_num)02d',
            'command': command,
            'user': storage.get('config.deploy.user'),
            'numprocs': str(numprocs),
            'storage': storage.get,
            'paths': paths.get
        })

        items.append(item)

    path_supervisor = '{}/supervisor/supervisor.j2'.format(
        paths.get('catapult.templates'),
    )

    generated = render_path(path_supervisor, {
        'items': '\n\n'.join(map(str, items)),
        'storage': storage.get,
        'paths': paths.get
    })

    return generated



import importlib
from catapult.deploy.ssh import Ssh
from catapult.deploy.paths import Paths
from catapult.deploy.notifier import Notifier
from catapult.deploy.recipes import Recipes
from catapult.deploy.storage import Storage
from catapult.deploy.locker import Locker
from catapult.deploy.blocker import Blocker
from catapult.deploy.constants import HOST_ALL, PHASES
from catapult.deploy.logger import Logger
from catapult.library.helpers import get_place
from catapult.library.exceptions import BlockerException
from catapult.deploy.constants import STATE_BEFORE, STATE_AFTER


def factory(project_path, config, services):
    def deploy(request, logger):
        logger = Logger(logger)

        try:
            place, storage, ssh, paths, notifier, locker, blocker, recipes = get_instances(
                project_path,
                config,
                services,
                request,
                logger
            )
        except Exception as error:
            logger.error('An error occurred with initializing: {}'.format(error))

            return False

        try:
            for phase in PHASES:
                logger.phase(
                    phase,
                    request['service'],
                    request['branch'],
                    request['place'],
                )

                recipes.execute(STATE_BEFORE, phase)

                imported_module = importlib.import_module('catapult.deploy.phases.{}'.format(phase))

                imported_module.phase(
                    paths,
                    storage,
                    locker,
                    blocker,
                    ssh,
                    logger
                )

                recipes.execute(STATE_AFTER, phase)

            notifier.notify(True)
        except Exception as error:
            logger.error(error)

            # release local lock
            if not isinstance(error, BlockerException):
                blocker.release_lock()

            try:
                notifier.notify(False)
            except Exception as error:
                logger.error('an error occurred with notifying: "{}"'.format(error))

            return False

        return True

    return deploy


def get_instances(project_path, config, services, request, logger):
    place = get_place(
        services,
        request['service'],
        request['place'],
    )

    storage = Storage(
        config,
        place,
        request
    )

    ssh = Ssh(
        config['deploy']['ssh']['user'],
        config['deploy']['ssh']['key_file'],
        storage.get_hosts(HOST_ALL),
        logger
    )

    paths = Paths(
        project_path,
        request['service'],
        config['deploy']['deploy_path'],
        request['name'],
    )

    notifier = Notifier(
        place['notifications'],
        request,
        logger
    )

    locker = Locker(
        paths,
        storage,
        ssh,
        logger
    )

    blocker = Blocker(
        paths.get('local.releases_folder'),
        request['place'],
        request['name']
    )

    recipes = Recipes(
        storage,
        paths,
        ssh,
        logger
    )

    return place, storage, ssh, paths, notifier, locker, blocker, recipes



from os import makedirs
from catapult.deploy.constants import SERVER_WEB, SERVER_BOT, SERVER_MAINTAIN


def phase(paths, storage, _, __, ___, logger):
    logger.success('release started by name "{}"'.format(storage.get('release.name')))
    makedirs(paths.get('local.code'))

    place = storage.get('place')

    # show list of recipes
    if place['recipes']:
        for recipe in place['recipes']:
            logger.info('recipe "{}" will be used'.format(recipe))

    # show list of servers
    if 'cloud' in place:
        logger.info('used "cloud mode" for deploying')

        for host_type in (SERVER_WEB, SERVER_BOT, SERVER_MAINTAIN):
            if host_type in place['cloud']:
                for item in place['cloud'][host_type]:
                    logger.info(
                        '"{}" host "{}" will be used'.format(
                            host_type,
                            item['host']
                        )
                    )
    else:
        logger.info('used "cluster mode" for deploying')

        for server in place['cluster']:
            logger.info(
                'host "{}" will be used'.format(
                    server['host']
                )
            )

    # show list of modules
    if place['modules']:
        for module in place['modules']:
            logger.info('module "{}" will be used'.format(module))



from catapult.library.renders import render_path, render_string
from catapult.modules.normalizers import normalize_systemd
from catapult.deploy.constants import HOST_ALL
from catapult.deploy.constants import REGISTER_CHECK, REGISTER_INSTALL, REGISTER_WEB_AND_BOT


class Module:
    def __init__(self, module_config, storage, paths, register, logger):
        self.module_config = module_config
        self.storage = storage
        self.paths = paths
        self.logger = logger

        register(REGISTER_INSTALL, REGISTER_WEB_AND_BOT, self.render_install())
        register(REGISTER_CHECK, REGISTER_WEB_AND_BOT, self.render_check())

    def render_check(self):
        path = '{}/systemd/check.j2'.format(
            self.paths.get('catapult.templates'),
        )

        generated = render_path(path, {
            'storage': self.storage.get,
            'paths': self.paths.get,
        })

        return generated

    def render_install(self):
        path = '{}/systemd/install.j2'.format(
            self.paths.get('catapult.templates'),
        )

        generated = render_path(path, {
            'release_path': self.paths.get('remote.release'),
            'service_name': 'catapult-{}'.format(self.storage.get('release.service')),
            'storage': self.storage.get,
            'paths': self.paths.get,
        })

        return generated

    def configure(self, config):
        config = normalize_systemd(config)
        hosts = self.storage.get_hosts(HOST_ALL)

        self.logger.info('configuring systemd unit file for all hosts: {}'.format(
            ', '.join(hosts))
        )

        configs = {}
        for host in hosts:
            self.storage.start_host(host)

            configs[host] = render(
                config,
                self.storage,
                self.paths
            )

            self.storage.finish_host()

        return configs


def render(config, storage, paths):
    path_environment = '{}/systemd/environment.j2'.format(
        paths.get('catapult.templates'),
    )

    path_unit = '{}/systemd/unit.j2'.format(
        paths.get('catapult.templates'),
    )

    environments = []
    for key, value in config['environments'].items():
        generated = render_path(path_environment, {
            'key': key,
            'value': value
        })

        environments.append(generated)

    execute = render_string(config['execute'], {
        'release_path': paths.get('remote.code'),
        'storage': storage.get,
        'paths': paths.get,
    })

    generated = render_path(path_unit, {
        'name': storage.get('release.service'),
        'user': storage.get('config.deploy.user'),
        'exec_start': execute,
        'environments': '\n'.join(environments),
    })

    return generated



import re
import time
from catapult.deploy.constants import LOGGER_ERROR, LOGGER_INFO, LOGGER_PHASE, LOGGER_SUCCESS, LOGGER_WARNING


class Logger:
    def __init__(self, logger):
        self.logger = logger

    def info(self, message, host=None):
        self.logger.event(LOGGER_INFO, draw(message, host))

    def error(self, message, host=None):
        self.logger.event(LOGGER_ERROR, draw(cut_message(message), host))

    def success(self, message, host=None):
        self.logger.event(LOGGER_SUCCESS, draw(cut_message(message), host))

    def warning(self, message, host=None):
        self.logger.event(LOGGER_WARNING, draw(message, host))

    def phase(self, phase, service, branch, place):
        message = '{} [service: {} | branch: {} | place: {}]'.format(
            phase.upper(),
            service.lower(),
            branch.lower(),
            place.lower()
        )

        self.logger.event(LOGGER_PHASE, draw(message + ' ' + '*' * (150 - len(message)), None))


def draw(message, host):
    if host:
        message = '[{}] {}'.format(host, message)

    message = '[{}] {}'.format(time.strftime('%H:%M:%S'), message)

    return message


def cut_message(message):
    message = str(message)

    message = re.sub('\u001b\[[0-9]*[A-K]', '', message)
    message = re.sub('\u001b\[[0-9;]*m', '', message)

    message = message.strip()

    if len(message) > 250:
        return message[0: 250] + '...'

    return message.strip()



from catapult.library.renders import render_path, render_string
from catapult.library.helpers import get_hosts
from catapult.modules.balancer import balanced_by_host
from catapult.modules.normalizers import normalize_crons
from catapult.deploy.constants import SERVER_BOT
from catapult.deploy.constants import REGISTER_CHECK, REGISTER_INSTALL, REGISTER_BOT


class Module:
    def __init__(self, module_config, storage, paths, register, logger):
        self.module_config = module_config
        self.storage = storage
        self.paths = paths
        self.logger = logger

        register(REGISTER_CHECK, REGISTER_BOT, self.render_check_bot())
        register(REGISTER_INSTALL, REGISTER_BOT, self.render_install_bot())

    def render_check_bot(self):
        path = '{}/crontab/check.j2'.format(
            self.paths.get('catapult.templates'),
        )

        generated = render_path(path, {
            'storage': self.storage.get,
            'paths': self.paths.get,
        })

        return generated

    def render_install_bot(self):
        path = '{}/crontab/install.j2'.format(
            self.paths.get('catapult.templates'),
        )

        generated = render_path(path, {
            'release_path': self.paths.get('remote.release'),
            'file_name': 'catapult-{}'.format(self.storage.get('release.service')),
            'storage': self.storage.get,
            'paths': self.paths.get
        })

        return generated

    def configure(self, config):
        crons = normalize_crons(config)
        servers = self.storage.get_servers(SERVER_BOT)

        if not servers or not config:
            self.logger.warning('there are not bot servers')

            return

        self.logger.info('configuring bot hosts: {}'.format(
            ', '.join(get_hosts(servers)))
        )

        crontab = get_crontab(self.storage, crons, servers)

        # render
        configs = {}
        for host in crontab:
            self.storage.start_host(host)

            configs[host] = render(
                crontab[host],
                self.storage,
                self.paths
            )

            self.storage.finish_host()

        return configs


def get_crontab(storage, crons, servers):
    crontab = {}

    # cluster mode
    if storage.is_cluster_mode():
        for server in servers:
            crontab[server['host']] = crons

        return crontab

    # cloud mode
    balancing = balanced_by_host(crons, servers)

    for server in servers:
        crontab[server['host']] = []

        for cron in balancing[server['host']]:
            crontab[server['host']].append(cron)

    return crontab


def render(crons, storage, paths):
    items = []

    path_item = '{}/crontab/item.j2'.format(
        paths.get('catapult.templates'),
    )

    for cron in crons:
        command = render_string(cron['command'], {
            'release_path': paths.get('remote.code'),
            'storage': storage.get,
            'paths': paths.get,
        })

        item = render_path(path_item, {
            'minutes': str(cron['planning']['minutes']),
            'hours': str(cron['planning']['hours']),
            'dom': str(cron['planning']['dom']),
            'mon': str(cron['planning']['mon']),
            'dow': str(cron['planning']['dow']),
            'command': command,
            'user': storage.get('config.deploy.user'),
            'storage': storage.get,
            'paths': paths.get,
        })

        items.append(item)

    path_crontab = '{}/crontab/crontab.j2'.format(
        paths.get('catapult.templates'),
    )

    generated = render_path(path_crontab, {
        'items': '\n\n'.join(map(str, items)),
        'storage': storage.get,
        'paths': paths.get,
    })

    return generated



#!/usr/bin/env python3

from setuptools import setup, find_packages
from catapult.version import __version__


def requirements():
    requirements_list = []

    with open('requirements.txt') as requirements:
        for install in requirements:
            requirements_list.append(install.strip())

    return requirements_list


packages = find_packages(exclude=['tests*'])

setup(name='catapult3',
      version=__version__,
      author='Artemiy Pulyavin',
      author_email='artemiy@pulyavin.ru',
      license='LGPLv3',
      keywords='deploy',
      description='Catapult - DevOps Deployment System',
      packages=packages,
      install_requires=requirements(),
      include_package_data=True,
      scripts=['bin/catapult'],
      zip_safe=False,
      )



import os
import re
import yaml
import shutil
import logging
import random
from datetime import datetime
from catapult.library.exceptions import ConfigParserException, InvalidParameterException
from catapult.deploy.constants import CLOUD, CLUSTER
from catapult.deploy.constants import SERVER_WEB, SERVER_BOT, SERVER_MAINTAIN


def parse_yaml(path):
    if not os.path.isfile(path):
        raise ConfigParserException('Yaml file by path "{}" does not found'.format(path))

    with open(path) as stream:
        try:
            return yaml.load(stream, Loader=yaml.Loader)
        except Exception as error:
            raise ConfigParserException(
                'An error occurred with parsing yaml file by path "{}":\n\n{}'.format(
                    path,
                    indented_error(error)
                )
            )


def indented_error(error):
    indent = '>>> '
    splitted = str(error).split('\n')

    return indent + ('\n' + indent).join(splitted)


def generate_release_name(service, branch):
    if len(branch) > 25:
        branch = branch[0:25]

    time_start = str('{:%y%m%d%H%M%S}'.format(datetime.now()))

    number = int(random.uniform(0, 20))
    if number <= 9:
        number = '0' + str(number)
    number = str(number)

    cleaned_branch = re.sub('[^A-Za-z0-9]+', '_', branch)
    cleaned_service = re.sub('[^A-Za-z0-9]+', '_', service)

    release_name = '{}{}_{}_{}'.format(
        time_start,
        number,
        cleaned_service,
        cleaned_branch
    )

    return release_name.lower()


def get_nested(path, dictionary):
    splitted = path.split('.')

    for section in splitted:
        if section not in dictionary:
            raise InvalidParameterException(
                'Section with name "{}" does not found in path "{}"'.format(section, path)
            )

        dictionary = dictionary[section]

        if dictionary is None:
            raise InvalidParameterException(
                'Section with name "{}" is empty in path "{}"'.format(section, path)
            )

    return dictionary


def get_place(services, service, place):
    if service not in services:
        raise ConfigParserException('Service "{}" does not found'.format(service))

    if place not in services[service]['places']:
        raise ConfigParserException('Place "{}" does not found in service "{}"'.format(place, service))

    return services[service]['places'][place]


def normalize_linked_paths(paths):
    if paths is None:
        return []

    paths = sorted(paths, key=len, reverse=True)

    normalized = []

    for path in paths:
        normalized.append(path.strip('/'))

    linked_paths = '"{0}"'.format('" "'.join(normalized))

    return linked_paths


def get_hosts(servers):
    hosts = []

    for server in servers:
        hosts.append(server['host'])

    return hosts


def get_messages(main_transport_config):
    message_success = 'Branch "{{ branch }}" of service "{{ service }}" ' \
                      'successfully deployed for place "{{ place }}" with {{ time }}s'
    message_fail = 'Error with deploy branch "{{ branch }}" of service "{{ service }}" for place "{{ place }}"'

    if 'messages' not in main_transport_config:
        return {
            'success': message_success,
            'fail': message_fail,
        }

    if 'success' in main_transport_config['messages']:
        message_success = main_transport_config['messages']['success']

    if 'fail' in main_transport_config['messages']:
        message_fail = main_transport_config['messages']['fail']

    return {
        'success': message_success,
        'fail': message_fail,
    }


def remove_releases_folder(project_path):
    releases_path = os.path.join(project_path, 'releases')

    if os.path.exists(releases_path):
        logging.info('deleting folder {}'.format(releases_path))

        shutil.rmtree(releases_path)


def start_logging(project_path):
    logging_path = os.path.join(project_path, 'server.log')

    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s [%(levelname)-5.5s] %(message)s",
        filename=logging_path,
    )

    logging.info('-' * 100)


def parse_nested_parameters(parameters, prefix=''):
    reducer = []

    for key, value in parameters.items():
        prefixed = prefix
        if prefix:
            prefixed = '{}.'.format(prefix)

        name = '{}{}'.format(prefixed, key)

        if isinstance(value, dict):
            reducer += parse_nested_parameters(value, name)
        else:
            reducer.append(name)

    return reducer


def get_parameters_by_host(place):
    parameters = {}

    if CLOUD in place:
        for host_type in (SERVER_WEB, SERVER_BOT, SERVER_MAINTAIN):
            if host_type in place[CLOUD]:
                for server in place[CLOUD][host_type]:
                    parameters[server['host']] = server['parameters']

    if CLUSTER in place:
        for server in place[CLUSTER]:
            parameters[server['host']] = server['parameters']

    return parameters



LOGGER_INFO = 'info'
LOGGER_ERROR = 'error'
LOGGER_SUCCESS = 'success'
LOGGER_WARNING = 'warning'
LOGGER_PHASE = 'phase'

CLUSTER = 'cluster'
CLOUD = 'cloud'

STATE_BEFORE = 'before'
STATE_AFTER = 'after'

TRANSPORT_TELEGRAM = 'telegram'
TRANSPORT_SLACK = 'slack'
TRANSPORT_PUSHOVER = 'pushover'

BULLET_LOCAL = 'local'
BULLET_REMOTE = 'remote'
BULLET_MAINTAIN = 'maintain'
BULLET_CONSOLE = 'console'

SERVER_WEB = 'web'
SERVER_BOT = 'bot'
SERVER_MAINTAIN = 'maintain'

HOST_ALL = 'all'
HOST_WEB = 'web'
HOST_BOT = 'bot'
HOST_MAINTAIN = 'maintain'

REGISTER_WEB = 'web'
REGISTER_BOT = 'bot'
REGISTER_WEB_AND_BOT = 'web_and_bot'
REGISTER_CHECK = 'check'
REGISTER_INSTALL = 'install'

PHASES = [
    'starting',
    'locking',
    'configuring',
    'building',
    'checking',
    'releasing',
    'maintaining',
    'installing',
    'cleaning',
    'ending',
]



from os.path import exists
from os import makedirs, chmod
from importlib import import_module
from catapult.library.exceptions import ModuleException
from catapult.library.renders import render_path
from catapult.library.helpers import normalize_linked_paths
from catapult.deploy.constants import REGISTER_INSTALL, REGISTER_CHECK
from catapult.deploy.constants import REGISTER_WEB, REGISTER_BOT, REGISTER_WEB_AND_BOT


class Modules:
    def __init__(self, storage, paths, logger):
        self.storage = storage
        self.paths = paths
        self.logger = logger

        self.modules = {}
        self.registers = {}

        modules = {}
        if storage.get('place.modules'):
            modules = storage.get('place.modules')

        for module_name, module_config in modules.items():
            self.logger.info('instantiate module "{}"'.format(module_name))

            if module_config is None:
                self.logger.info('service config for module "{}" is empty'.format(module_name))

                module_config = {}

            path = '{}/{}/module.py'.format(
                paths.get('catapult.modules'),
                module_name
            )

            if not exists(path):
                raise ModuleException('Module "{}" does not found'.format(module_name))

            imported_module = import_module('catapult.modules.{}.module'.format(module_name))

            instance = imported_module.Module(
                module_config,
                self.storage,
                self.paths,
                self.register,
                self.logger
            )

            self.modules[module_name] = instance

    def register(self, stage, host_type, bash_code):
        if stage not in (REGISTER_INSTALL, REGISTER_CHECK):
            raise ModuleException('Stage type "{}" is not defined'.format(stage))

        if host_type not in (REGISTER_BOT, REGISTER_WEB, REGISTER_WEB_AND_BOT):
            raise ModuleException('Hosts type "{}" is not defined'.format(host_type))

        if stage not in self.registers:
            self.registers[stage] = {}

        if host_type not in self.registers[stage]:
            self.registers[stage][host_type] = []

        self.registers[stage][host_type].append(bash_code)

    def configure(self, deploy_yaml):
        for module_name, module_instance in self.modules.items():
            self.logger.info('configure deploy.yml section "{}"'.format(module_name))

            if module_name not in deploy_yaml:
                self.logger.warning('section "{}" does not exist in deploy.yml, skipped'.format(module_name))

                continue

            config = deploy_yaml[module_name]

            config_by_host = module_instance.configure(config)

            if not type(config_by_host) is dict:
                self.logger.info('module "{}" does not return configure data, skipped'.format(module_name))
                continue

            for host, generated in config_by_host.items():
                folder = '{}/{}'.format(self.paths.get('local.configs'), host)

                if not exists(folder):
                    makedirs(folder)

                with open('{}/{}'.format(folder, module_name), 'w') as open_file:
                    open_file.write(generated)

        self.logger.info('configuring install scripts')
        configure_install(self.storage, self.paths, self.registers)

        self.logger.info('configuring check scripts')
        configure_check(self.storage, self.paths, self.registers)


def configure_install(storage, paths, registers):
    web_items = ''
    if REGISTER_INSTALL in registers and REGISTER_WEB in registers[REGISTER_INSTALL]:
        web_items = '\n'.join(registers[REGISTER_INSTALL][REGISTER_WEB])

    bot_items = ''
    if REGISTER_INSTALL in registers and REGISTER_BOT in registers[REGISTER_INSTALL]:
        bot_items = '\n'.join(registers[REGISTER_INSTALL][REGISTER_BOT])

    web_and_bot_items = ''
    if REGISTER_INSTALL in registers and REGISTER_WEB_AND_BOT in registers[REGISTER_INSTALL]:
        web_and_bot_items = '\n'.join(registers[REGISTER_INSTALL][REGISTER_WEB_AND_BOT])

    linked_paths = normalize_linked_paths(storage.get('place.linked'))

    path = '{}/core/install.j2'.format(
        paths.get('catapult.templates'),
    )

    generated = render_path(path, {
        'release_path': paths.get('remote.release'),
        'release_name': storage.get('release.name'),
        'user': storage.get('config.deploy.user'),
        'service_name': storage.get('release.service'),
        'file_name': 'catapult-{}'.format(storage.get('release.service')),
        'deploy_path': storage.get('config.deploy.deploy_path'),
        'linked_paths': linked_paths,
        'web_items': web_items,
        'bot_items': bot_items,
        'web_and_bot_items': web_and_bot_items,
    })

    file = '{}/install.sh'.format(paths.get('local.release'))
    with open(file, 'w') as open_file:
        open_file.write(generated)

    chmod(file, 0o0766)


def configure_check(storage, paths, registers):
    web_items = ''
    if REGISTER_CHECK in registers and REGISTER_WEB in registers[REGISTER_CHECK]:
        web_items = '\n'.join(registers[REGISTER_CHECK][REGISTER_WEB])

    bot_items = ''
    if REGISTER_CHECK in registers and REGISTER_BOT in registers[REGISTER_CHECK]:
        bot_items = '\n'.join(registers[REGISTER_CHECK][REGISTER_BOT])

    web_and_bot_items = ''
    if REGISTER_CHECK in registers and REGISTER_WEB_AND_BOT in registers[REGISTER_CHECK]:
        web_and_bot_items = '\n'.join(registers[REGISTER_CHECK][REGISTER_WEB_AND_BOT])

    path = '{}/core/check.j2'.format(
        paths.get('catapult.templates'),
    )

    generated = render_path(path, {
        'user': storage.get('config.deploy.user'),
        'service_name': storage.get('release.service'),
        'deploy_path': storage.get('config.deploy.deploy_path'),
        'web_items': web_items,
        'bot_items': bot_items,
        'web_and_bot_items': web_and_bot_items,
    })

    file = '{}/check.sh'.format(paths.get('local.release'))
    with open(file, 'w') as open_file:
        open_file.write(generated)

    chmod(file, 0o0766)



import tempfile
import logging
from time import sleep
from multiprocessing import Queue, Process, active_children
from catapult.server.processes.worker import worker
from catapult.server.processes.writer import writer
from catapult.server.processes.server import server
from catapult.server.library import get_ws_url
from catapult.server.killer import GracefulKiller, GracefulKillerException
from catapult.server.repository import Repository

processes = []


def master(config, services, deploy):
    sqlite_path = '{}/base.sqlite'.format(tempfile.mkdtemp())
    logging.info('use path "{}" for sqlite'.format(sqlite_path))

    deploy_queue = Queue()
    write_queue = Queue()

    repository = Repository(sqlite_path)
    repository.initialize()

    workers = config['server']['workers']
    ws_template = get_ws_url(
        config['server']['host'],
        config['server']['port']
    )

    logging.info('run {} deploy workers'.format(workers))
    for _ in range(0, workers):
        arguments = (
            deploy,
            deploy_queue,
            write_queue,
            ws_template
        )
        start_process(worker, arguments)

    logging.info('run writer process')
    start_process(writer, (
        repository,
        write_queue,
    ))

    logging.info('run server process')
    start_process(server, (
        config,
        services,
        repository,
        deploy_queue,
        write_queue,
    ))

    killer = GracefulKiller()

    try:
        while len(processes) > 0:
            for index, structure in enumerate(processes):
                process = structure['process']
                target = structure['target']

                if not process.is_alive():
                    logging.error('process "{}" dying, try ro restart'.format(target.__name__))
                    restart_process(index)

            sleep(1)

            if killer.is_killed:
                break
    except GracefulKillerException:
        logging.info('master process shutting down')

    active_children()


def start_process(target, arguments):
    process = Process(target=target, args=arguments)
    process.daemon = False
    process.start()

    processes.append({
        'process': process,
        'target': target,
        'arguments': arguments
    })


def restart_process(index):
    structure = processes[index]

    process = Process(
        target=structure['target'],
        args=structure['arguments']
    )
    process.daemon = False
    process.start()

    structure['process'] = process



from os.path import abspath, dirname, join
from catapult.library.helpers import get_nested
from catapult.library.exceptions import InvalidParameterException


class Paths:
    def __init__(self, project_path, service, deploy_path, release_name):
        # generated absolute path to catapult project
        catapult_folder = abspath(join(dirname(__file__), '../'))

        self.paths = {
            'catapult': {
                'templates': '{}/templates'.format(
                    catapult_folder
                ),
                'modules': '{}/modules'.format(
                    catapult_folder
                ),
            },
            'local': {
                'project_path': project_path,
                'recipes': '{}/recipes'.format(
                    project_path
                ),
                'services': '{}/services'.format(
                    project_path
                ),
                'templates': '{}/templates'.format(
                    project_path
                ),
                'releases_folder': '{}/releases/{}'.format(
                    project_path,
                    service
                ),
                'release': '{}/releases/{}/{}'.format(
                    project_path,
                    service,
                    release_name
                ),
                'configs': '{}/releases/{}/{}/configs'.format(
                    project_path,
                    service,
                    release_name,
                ),
                'code': '{}/releases/{}/{}/code'.format(
                    project_path,
                    service,
                    release_name,
                ),
                'builds': '{}/releases/{}/{}/builds'.format(
                    project_path,
                    service,
                    release_name,
                ),
                'shared': '{}/releases/{}/{}/shared'.format(
                    project_path,
                    service,
                    release_name,
                ),
            },

            'remote': {
                'service_folder': '{}/{}'.format(
                    deploy_path,
                    service
                ),
                'releases_folder': '{}/{}/releases'.format(
                    deploy_path,
                    service
                ),
                'release': '{}/{}/releases/{}'.format(
                    deploy_path,
                    service,
                    release_name
                ),
                'code': '{}/{}/releases/{}/code'.format(
                    deploy_path,
                    service,
                    release_name
                ),
            }
        }

    def get(self, path, default=None):
        """
            catapult.templates
            catapult.modules

            local.project_path
            local.recipes
            local.services
            local.templates
            local.releases_folder
            local.release
            local.configs
            local.code
            local.builds
            local.shared

            remote.service_folder
            remote.releases_folder
            remote.release
            remote.code
        """
        try:
            return get_nested(path, self.paths)
        except InvalidParameterException as error:
            if default is None:
                raise error

            return default
